##Created by Joseph Long 2016
##Modified to resemble DOS edlin by David Hunter 2021
##First python project
##Simple Text Editor

import os
import sys
try:
    from pydos_ui import input
except:
    pass

loaded = []
autosave = False
filename = ""
lineNum = 0

def chkPath(tstPath):
    validPath = True

    if tstPath == []:
        validPath = True
    else:

        savDir = os.getcwd()

        for path in tstPath:
            if path == "":
                os.chdir("/")

            elif os.getcwd() == "/" and path == "..":
                validPath = False
                break

            elif path == "." or path == "..":
                os.chdir(path)

            elif path in os.listdir() and (os.stat(path)[0] & (2**15) == 0):
                os.chdir(path)

            else:
                validPath = False
                break

        os.chdir(savDir)

    return(validPath)

def open_text(name):

    if name == "":
        name = input("Enter file name: ")

    aPath = name.split("/")
    newdir = aPath.pop(-1)

    text = []
    if chkPath(aPath) and newdir in os.listdir(name[0:len(name)-len(newdir)]) and os.stat(name)[0] & (2**15) != 0:
        f = open(name, "r")

        for line in f:
            text.append(line.replace("\n",""))

        f.close()
    else:
        print("Unable to open: "+name+". File not found.")
        name = ""

    return(name,text)

def proc_text(lineNum,func,strt,end,text):
    if len(text) > 0:
        x = min(max(0,strt),len(text)-1)
        popIndx = x
        staticEnd = max(0,min(end,len(text)-1))
        if x == staticEnd:
            lineNum = staticEnd
        while x <= staticEnd:
            if func == "L":
                if x == lineNum:
                    print("   *", x, text[x])
                else:
                    print("    ", x, text[x])
            elif func == "D":
                text.pop(popIndx)
                lineNum = popIndx
            x+=1

    return(lineNum,text)

def save_text(name,text):
    if name == "":
        name = input("Enter file name: ")

    f = open(name, "w")

    for line in text:
        f.write(line + "\n")

    f.close()

    return(name)

def parseInput(command):

    global parseMap

    cmdList = []
    parseMap = "C"
    cmdFound = False
    cnfrmFound = False
    buildingNum = False
    builtNum =""
    txtParam = ""
    for sChar in command:
        if not cmdFound:
            if sChar not in "0123456789,.+-$?":
                cmd = sChar.upper()
                cmdFound = True
                if buildingNum:
                    cmdList.append(int(builtNum))
                    buildingNum = False
                    builtNum = ""
                    parseMap += "-"
            elif sChar in "0123456789+-":
                builtNum += sChar
                buildingNum = True
            elif sChar == "?" and not cmdFound:
                if not cnfrmFound:
                    parseMap += sChar
                cnfrmFound = True
            else:
                if buildingNum:
                    cmdList.append(int(builtNum))
                    buildingNum = False
                    builtNum = ""
                    parseMap += "-"
        else:
            txtParam += sChar

    if txtParam != "":
        cmdList.append(txtParam.strip())
        parseMap += "+"

    if not cmdFound:
        if buildingNum:
            cmdList.append(int(builtNum))
            parseMap += "-"
            cmd = "*"
        else:
            cmd = "#"

    cmdList.insert(0,cmd)
    return cmdList

def interperit(command,text):
    global filename
    global autosave
    global lineNum
    global parseMap

    confirmCmd = False
    loop = True
    command_list = parseInput(command)
    #print (lineNum,command_list,parseMap)
    if "?" in parseMap:
        confirmCmd = True
        parseMap = parseMap.replace("?","")

    if command_list[0] == "O":
        if parseMap == "C+":
            (filename,text) = open_text(command_list[1])
        else:
            (filename,text) = open_text(filename)

    elif command_list[0] == "W":
        if parseMap == "C+":
            filename = save_text(command_list[1],text)
        else:
            filename = save_text(filename,text)

    elif command_list[0] == "E":
        if parseMap == "C+":
            filename = save_text(command_list[1],text)
        else:
            filename = save_text(filename,text)
        loop = False

    elif command_list[0] == "Q":
        loop = False

    elif command_list[0] == "autosave":
        if autosave:
            autosave = False
            print("\tAutosave disabled")
        else:
            autosave = True
            print("\tAutosave enabled")

    elif command_list[0] == "new":
        if len(command_list) > 1:
            filename = save_text(command_list[1][1:],text)
        else:
            filename = save_text(filename,text)

        text[:] = []

    elif command_list[0] == "H":
        print(" ")
        print(" ")
        print("h                      - DISPLAY COMMANDS")
        print("#                      - REPLACE A SINGLE LINE")
        print("[#][,#]d               - DELETE A BLOCK OF LINES")
        print("[#]i                   - INSERT")
        print("[#]a['str']            - APPEND LINES")
        print("[#][,#]l               - LIST LINES")
        print("[#][,#][?]r'str','str' - REPLACE STRING")
        print("[#][,#][?]s'str'       - SEARCH FOR STRING")
        print("o [filename]           - OPEN FILE")
        print("w [filename]           - WRITE FILE")
        print("e [filename]           - SAVE AND QUIT")
        print("q                      - QUIT")

    elif command_list[0] == "L" or command_list[0] == "D":
        if parseMap == "C-":
            (lineNum,text) = proc_text(lineNum,command_list[0],int(command_list[1]),int(command_list[1]),text)
        elif parseMap == "C--":
            (lineNum,text) = proc_text(lineNum,command_list[0],int(command_list[1]),int(command_list[2]),text)
        elif parseMap == "C":
            (lineNum,text) = proc_text(lineNum,command_list[0],0,len(text)-1,text)
        else:
            print("* Wrong command format. use: [#][,#]"+command_list[0].lower())

    elif command_list[0] == "I":
        if parseMap == "C-":
            text.insert(int(command_list[1]), input("."))
        elif parseMap == "C":
            text.insert(lineNum, input("."))
        else:
            print("* Wrong command format. use: [#]i")

    elif command_list[0] == "A":
        if parseMap == "C":
            text.append(input("."))
        elif parseMap == "C-":
            for a in range(0, int(command_list[1])):
                text.append(input("."))
        elif parseMap == "C+" or parseMap == "C-+":
            strngs = command_list[-1].strip()
            if strngs.count(strngs[0]) !=2:
                print("* Wrong command format. Use: [#]a 'str'")
            else:
                strngs = strngs.replace(strngs[0],"")
                if parseMap == "C+":
                    text.append(strngs)
                elif parseMap == "C-+":
                    for a in range(0, int(command_list[1])):
                        text.append(strngs)
        else:
            print("* Wrong command format, use: [#]a 'str'")


    elif command_list[0] == "find":
        for b in range(0, len(text)):
            if text[b].find(command_list[1][1:]) != -1:
                print("    ", b, text[b])

    elif command_list[0] == "R":
        strt = lineNum
        if parseMap == "C+":
            strt = lineNum
            end = lineNum
        elif parseMap == "C-+":
            strt = int(command_list[1])
            end = strt
        elif parseMap == "C--+":
            strt = int(command_list[1])
            end = int(command_list[2])
        else:
            print("* Wrong command format. Use: [#][,#][?]r 'str1','str2'")
            strt = -1

        if strt != -1:
            strngs = command_list[-1].strip()
            delim = strngs[0]
            if strngs.count(delim) != 4:
                print("* Wrong command format. Use: [#][,#][?]r 'str1','str2'")
            else:

                x = 0
                token = []
                delimFound = True
                for achar in strngs:
                    if achar == delim:
                        if delimFound:
                            strtTk = x+1
                        else:
                            endTk = x
                            token.append(strngs[strtTk:endTk])
                        delimFound = not delimFound
                    x += 1
                if confirmCmd:
                    print("for lines:",strt,"to",end,"- replace:",token[0],"with:",token[1],":")
                for b in range(strt, end+1):
                    if confirmCmd:
                        print(text[b].replace(token[0], token[1]))
                        if input("Confirm change (y/n): ")[0].upper() == "Y":
                            text[b] = text[b].replace(token[0], token[1])
                    else:
                        text[b] = text[b].replace(token[0], token[1])

    elif command_list[0] == "S":
        strt = lineNum
        if parseMap == "C+":
            strt = 0
            end = len(text) - 1
        elif parseMap == "C-+":
            strt = int(command_list[1])
            end = strt
        elif parseMap == "C--+":
            strt = int(command_list[1])
            end = int(command_list[2])
        else:
            print("* Wrong command format. Use: [#][,#][?]s 'str'")
            strt = -1

        if strt != -1:
            strngs = command_list[-1].strip()
            if strngs.count(strngs[0]) !=2:
                print("* Wrong command format. Use: [#][,#][?]s 'str'")
            else:
                strngs = strngs.replace(strngs[0],"")
                for b in range(strt, end+1):
                    if text[b].count(strngs) > 0:
                        lineNum = b
                        proc_text(lineNum,"L",b,b,text)
                        if not confirmCmd:
                            break
                        else:
                            if input("Confirm find (y/n): ")[0].upper() == "Y":
                                break

    elif command_list[0] == "load":
        try:
            f = open(command_list[1][1:] + ".py", "r")
            f.close
            loaded.append(command_list[1][1:])
        except: print("\tCouldn't load " + command_list[1] + ",\n\t\tPlugin does not exist")
        print(loaded)

    elif command_list[0] in "*#":
        if command_list[0] == "*":
            lineNum = min(len(text)-1,max(0,command_list[1]))
        proc_text(lineNum,"L",lineNum,lineNum,text)
        strngs = input(".")
        if strngs != "":
            text.insert(lineNum,strngs)
            if lineNum < len(text)-1:
                text.pop(lineNum+1)

        if lineNum < len(text):
            lineNum += 1

    elif command_list[0] in loaded:
        save_text("text.temp",text)
        args = " ".join(command_list[1:])
        os.system(".\\" + command_list[0] + ".py " + args)
        open("text.temp")
        os.remove("text.temp")

    return(loop,text)

def main(passedIn):
    global filename
    global autosave
    global lineNum
    global parseMap

    text = []

    if passedIn != "":
        (filename,text) = open_text(passedIn)

    print("h for command list")

    loop = True
    while loop:
        (loop,text) = interperit(input(filename+": "),text)
        #if autosave:
            #filename = save_text()

    del filename
    del autosave
    del lineNum
    del parseMap


if __name__ != "PyDOS":
    passedIn = ""

main(passedIn)



import machine
from time import sleep_ms

Enable = 0x04  # Set Entry Mode
RS = 0x01  # Register select bit
Width = 20


class lcd:
    # initializes objects and lcd
    def __init__(self, ADDR, I2C):
        self.i2c = I2C
        self.address = ADDR
        self.BKLIGHT = 0x08  # Set Backlight ON
        self.lcd_write(0x03)
        self.lcd_write(0x03)
        self.lcd_write(0x03)
        self.lcd_write(0x02)

        self.lcd_write(0x20 | 0x08 | 0x04 | 0x00)  # Set Functions 2Line,5x8,4Bit Mode
        self.lcd_write(0x08 | 0x04)  # Turn Display On
        self.lcd_write(0x01)  # Clear Screen
        self.lcd_write(0x04 | 0x02)  # Set Entry Mode Left -> Right
        sleep_ms(300)

    def lcd_strobe(self, data):
        self.i2c.writeto(self.address, bytes([data | Enable | self.BKLIGHT]))
        sleep_ms(1)
        self.i2c.writeto(self.address, bytes([(data & ~Enable) | self.BKLIGHT]))
        sleep_ms(1)

    def lcd_write_four_bits(self, data):
        self.i2c.writeto(self.address, bytes([data | self.BKLIGHT]))
        self.lcd_strobe(data)

    # write a command to lcd
    def lcd_write(self, cmd, RS=0):
        self.lcd_write_four_bits(RS | (cmd & 0xF0))
        self.lcd_write_four_bits(RS | ((cmd << 4) & 0xF0))

    def set_line(self, line, col=0):
        if line == 1:
            self.lcd_write(0x80 + col)
        if line == 2:
            self.lcd_write(0xC0 + col)
        if line == 3:
            self.lcd_write(0x94 + col)
        if line == 4:
            self.lcd_write(0xD4 + col)

    def lcd_print(self, string, line=0, col=0):
        self.set_line(line, col)
        i = 1
        for char in string:
            if (i > Width) & (line < 4):
                line = line + 1
                self.set_line(line, 0)
                i = 1
            if (i > Width) & (line == 4):
                break
            self.lcd_write(ord(char), RS)
            i = i + 1

    def lcd_off(self):
        self.lcd_write(0x08 | 0x00)

    def lcd_on(self):
        self.lcd_write(0x08 | 0x04)

    def lcd_clear(self):
        self.lcd_write(0x01)  # Clear Screen
        self.lcd_write(0x02)  # Set Home

    def lcd_backlight(self, on):
        if on:
            self.BKLIGHT = 0x08
        else:
            self.BKLIGHT = 0x00

        self.lcd_write(0)



from sys import implementation

print("This program will set the CircuitPython filesystem access mode AFTER the next")
print("Powercycle. A Control-D softboot will NOT reset the filesystem access mode.\n")

if implementation.name.upper() != "CIRCUITPYTHON":
    print("This program only works under CircuitPython.")
    print('This microcontroller is currently running: "'+implementation.name+'"')
else:

    if __name__ != "PyDOS":
        passedIn = ""

    if passedIn == "":
        passedIn = input("Enter RO or RW: ")

    if passedIn.upper() != "RO" and passedIn.upper() != "RW":
        print ("Invalid Option")
    else:
        if passedIn.upper() == "RW":
            readOnly = "False"
        else:
            readOnly = "True"

        try:
            f = open('/boot.py','w')
            f.write('import storage'+"\n")
            f.write('storage.remount("/",'+readOnly+")\n")
            f.close()
        except:
            print("The filesystem currently appears to be readonly, you should be able to access")
            print("the flash storage as a mounted USB drive on your host system. To switch the")
            print("filesystem back to Read/Write, copy the boot.rw file to boot.py, safely")
            print("dismount the USB drive from the host and powercycle the microcontroller board.")



#
# Front-end for Micropython standard console IO
#
try:
    import usys as sys
except:
    import sys

class IO_DEVICE:
    def __init__(self):
        try:
            from micropython import kbd_intr
            kbd_intr(-1)
        except ImportError:
            pass
        if hasattr(sys.stdin, "buffer"):
            self.rd_raw_fct = sys.stdin.buffer.read
        else:
            self.rd_raw_fct = sys.stdin.read

    def wr(self, s):
        sys.stdout.write(s)

    def rd(self):
        return sys.stdin.read(1)

    def rd_raw(self):
        return self.rd_raw_fct(1)

    def deinit_tty(self):
        try:
            from micropython import kbd_intr
            kbd_intr(3)
        except ImportError:
            pass

    def get_screen_size(self):
        self.wr('\x1b[999;999H\x1b[6n')
        pos = ''
        char = self.rd() ## expect ESC[yyy;xxxR
        while char != 'R':
            pos += char
            char = self.rd()
        return [int(i, 10) for i in pos.lstrip("\n\x1b[").split(';')]

## test, if the Editor class is already present
if "pye_edit" not in globals().keys():
    from pye import pye_edit

def pye(*args, tab_size=4, undo=50):
    io_device = IO_DEVICE()
    ret = pye_edit(args, tab_size=tab_size, undo=undo, io_device=io_device)
    io_device.deinit_tty()
    return ret



from spidevice import SPIDevice
from machine import Pin
import time

_SET_PIN_MODE_CMD = const(0x50)
_SET_DIGITAL_WRITE_CMD = const(0x51)
_SET_ANALOG_WRITE_CMD = const(0x52)
_SET_DIGITAL_READ_CMD = const(0x53)
_SET_ANALOG_READ_CMD = const(0x54)

_START_CMD = const(0xE0)
_END_CMD = const(0xEE)
_ERR_CMD = const(0xEF)
_REPLY_FLAG = const(1 << 7)
_CMD_FLAG = const(0)

class ESP_SPIcontrol:

    def __init__(self, spi, cs_dio, ready_dio, reset_dio, gpio0_dio=None, *, debug=False):
        self._debug = debug
        self.set_psk = False
        self.set_crt = False
        self._buffer = bytearray(10)
        self._pbuf = bytearray(1)  # buffer for param read
        self._sendbuf = bytearray(256)  # buffer for command sending
        self._socknum_ll = [[0]]  # pre-made list of list of socket #

        self._spi_device = SPIDevice(spi, cs_dio, baudrate=8000000)
        self._cs = cs_dio
        self._ready = ready_dio
        self._reset = reset_dio
        self._gpio0 = gpio0_dio
        #self._cs.direction = Pin.OUT
        #self._ready.direction = Pin.IN
        #self._reset.direction = Pin.OUT
        # Only one TLS socket at a time is supported so track when we already have one.
        self._tls_socket = None
        #if self._gpio0:
            #self._gpio0.direction = Pin.IN
        self.reset()

    def _send_command(self, cmd, params=None, *, param_len_16=False):
        """Send over a command with a list of parameters"""
        if not params:
            params = ()

        packet_len = 4  # header + end byte
        for i, param in enumerate(params):
            packet_len += len(param)  # parameter
            packet_len += 1  # size byte
            if param_len_16:
                packet_len += 1  # 2 of em here!
        while packet_len % 4 != 0:
            packet_len += 1
        # we may need more space
        if packet_len > len(self._sendbuf):
            self._sendbuf = bytearray(packet_len)

        self._sendbuf[0] = _START_CMD
        self._sendbuf[1] = cmd & ~_REPLY_FLAG
        self._sendbuf[2] = len(params)

        # handle parameters here
        ptr = 3
        for i, param in enumerate(params):
            if self._debug >= 2:
                print("\tSending param #%d is %d bytes long" % (i, len(param)))
            if param_len_16:
                self._sendbuf[ptr] = (len(param) >> 8) & 0xFF
                ptr += 1
            self._sendbuf[ptr] = len(param) & 0xFF
            ptr += 1
            for j, par in enumerate(param):
                self._sendbuf[ptr + j] = par
            ptr += len(param)
        self._sendbuf[ptr] = _END_CMD

        self._wait_for_ready()
        with self._spi_device as spi:
            times = time.ticks_ms()
            while (time.ticks_ms() - times) < 1000:  # wait up to 1000ms
                if self._ready.value():  # ok ready to send!
                    break
            else:
                raise RuntimeError("ESP32 timed out on SPI select")
            spi.write(self._sendbuf)
            # , start=0, end=packet_len)  # pylint: disable=no-member
            if self._debug >= 3:
                print("Wrote: ", [hex(b) for b in self._sendbuf[0:packet_len]])

    def _send_command_get_response(
        self,
        cmd,
        params=None,
        *,
        reply_params=1,
        sent_param_len_16=False,
        recv_param_len_16=False
    ):
        """Send a high level SPI command, wait and return the response"""
        self._send_command(cmd, params, param_len_16=sent_param_len_16)
        return self._wait_response_cmd(cmd, reply_params, param_len_16=recv_param_len_16)

    def _wait_response_cmd(self, cmd, num_responses=None, *, param_len_16=False):
        """Wait for ready, then parse the response"""
        self._wait_for_ready()

        responses = []
        with self._spi_device as spi:
            times = time.ticks_ms()
            while (time.ticks_ms() - times) < 1000:  # wait up to 1000ms
                if self._ready.value():  # ok ready to send!
                    break
            else:
                raise RuntimeError("ESP32 timed out on SPI select")

            self._wait_spi_char(spi, _START_CMD)
            self._check_data(spi, cmd | _REPLY_FLAG)
            if num_responses is not None:
                self._check_data(spi, num_responses)
            else:
                num_responses = self._read_byte(spi)
            for num in range(num_responses):
                param_len = self._read_byte(spi)
                if param_len_16:
                    param_len <<= 8
                    param_len |= self._read_byte(spi)
                if self._debug >= 2:
                    print("\tParameter #%d length is %d" % (num, param_len))
                response = bytearray(param_len)
                self._read_bytes(spi, response)
                responses.append(response)
            self._check_data(spi, _END_CMD)

        if self._debug >= 2:
            print("Read %d: " % len(responses[0]), responses)
        return responses

    def _wait_spi_char(self, spi, desired):
        """Read a byte with a retry loop, and if we get it, check that its what we expect"""
        for _ in range(10):
            r = self._read_byte(spi)
            if r == _ERR_CMD:
                raise RuntimeError("Error response to command")
            if r == desired:
                return True
            time.sleep(0.01)
        raise RuntimeError("Timed out waiting for SPI char")

    def _read_byte(self, spi):
        """Read one byte from SPI"""
        spi.readinto(self._pbuf)
        if self._debug >= 3:
            print("\t\tRead:", hex(self._pbuf[0]))
        return self._pbuf[0]

    def _read_bytes(self, spi, buffer, start=0, end=None):
        """Read many bytes from SPI"""
        if not end:
            end = len(buffer)
        #spi.readinto(buffer, start=start, end=end)
        spi.readinto(buffer)
        if self._debug >= 3:
            print("\t\tRead:", [hex(i) for i in buffer])

    def _check_data(self, spi, desired):
        """Read a byte and verify its the value we want"""
        r = self._read_byte(spi)
        if r != desired:
            raise RuntimeError("Expected %02X but got %02X" % (desired, r))

    def set_pin_mode(self, pin, mode):
        """Set the io mode for a GPIO pin.
        :param int pin: ESP32 GPIO pin to set.
        :param value: direction for pin, digitalio.Direction or integer (0=input, 1=output).
        """
        #if mode == Direction.OUTPUT:
            #pin_mode = 1
        #elif mode == Direction.INPUT:
            #pin_mode = 0
        #else:
            #pin_mode = mode
        pin_mode = mode
        resp = self._send_command_get_response(_SET_PIN_MODE_CMD, ((pin,), (pin_mode,)))
        if resp[0][0] != 1:
            raise RuntimeError("Failed to set pin mode")

    def set_analog_write(self, pin, analog_value):
        """Set the analog output value of pin, using PWM.
        :param int pin: ESP32 GPIO pin to write to.
        :param float value: 0=off 1.0=full on
        """
        value = int(255 * analog_value)
        resp = self._send_command_get_response(
            _SET_ANALOG_WRITE_CMD, ((pin,), (value,))
        )
        if resp[0][0] != 1:
            raise RuntimeError("Failed to write to pin")

    def reset(self):
        """Hard reset the ESP32 using the reset pin"""
        if self._debug:
            print("Reset ESP32")
        if self._gpio0:
            self._gpio0.direction = Direction.OUTPUT
            self._gpio0.value(True)  # not bootload mode
        self._cs.value(True)
        self._reset.value(False)
        time.sleep(0.01)  # reset
        self._reset.value(True)
        time.sleep(0.75)  # wait for it to boot up
        #if self._gpio0:
            #self._gpio0.direction = Direction.INPUT


    def _wait_for_ready(self):
        """Wait until the ready pin goes low"""
        if self._debug >= 3:
            print("Wait for ESP32 ready", end="")
        times = time.ticks_ms()
        while (time.ticks_ms() - times) < 10000:  # wait up to 10 seconds
            if not self._ready.value():  # we're ready!
                break
            if self._debug >= 3:
                print(".", end="")
                time.sleep(0.05)
        #else:
            #raise RuntimeError("ESP32 not responding")
        if self._debug >= 3:
            print()



import os
from time import localtime
from sys import stdin,implementation
try:
    from pydos_ui import Pydos_ui
except ImportError:
    Pydos_ui = None
try:
    from pydos_ui import input
except ImportError:
    pass

import gc
if implementation.name.upper() == "MICROPYTHON":
    from micropython import mem_info
elif implementation.name.upper() == "CIRCUITPYTHON":
    from supervisor import runtime

gc.collect()
if 'threshold' in dir(gc):
    gc.threshold(gc.mem_free() // 4 + gc.mem_alloc())



# The first string may contain wildcard characters
def _match(first, second):

    # If we reach at the end of both strings, we are done
    if len(first) == 0 and len(second) == 0:
        return True

    # Make sure that the characters after '*' are present
    # in second string. first string will can't contain two consecutive '*'
    if len(first) > 1 and first[0] == '*' and  len(second) == 0:
        return False

    if (len(first) > 1 and first[0] == '?') or (len(first) != 0
        and len(second) !=0 and first[0] == second[0]):
        return _match(first[1:],second[1:]);

    if len(first) !=0 and first[0] == '*':
        return _match(first[1:],second) or _match(first,second[1:])

    return False

def calcWildCardLen(wildcardLen,recursiveFail):
    wildcardLen += 1
    if not recursiveFail and wildcardLen < 90:
        try:
            (wildcardLen,recursiveFail) = calcWildCardLen(wildcardLen,recursiveFail)
        except:
            recursiveFail = True

    return (wildcardLen,recursiveFail)

def PyDOS():

    global envVars
    envVars = {}
    if implementation.name.upper() == "CPYTHON":
        slh = '\\'
    else:
        slh = '/'

    print("Starting Py-DOS...")
    if Pydos_ui:
        (envVars["_scrHeight"],envVars["_scrWidth"]) = Pydos_ui.get_screensize()
    else:
        envVars["_scrHeight"] = 24
        envVars["_scrWidth"] = 80

    wildcardLen = 0
    recursiveFail = False

    (wildcardLen,recursiveFail) = calcWildCardLen(wildcardLen,recursiveFail)

    wildcardLen = max(18,wildcardLen) - 2
    if wildcardLen < 40:
        print("*Warning* wild card length set to: ",wildcardLen)
    gc.collect()

    def anyKey():
        print("Press any key to continue . . . .",end="")
        if Pydos_ui:
            while not Pydos_ui.serial_bytes_available():
                pass
            keyIn = Pydos_ui.read_keyboard(1)
        else:
            if implementation.name.upper() == "CIRCUITPYTHON":
                while not runtime.serial_bytes_available:
                    pass
            keyIn = stdin.read(1)
        print("")
        return(keyIn)

    def weekDay():
        offset = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334]
        week   = ['Sun','Mon','Tue','Wed','Thu','Fri','Sat']
        month = localtime()[1]
        day = localtime()[2]
        year = localtime()[0]
        afterFeb = 1
        if month > 2: afterFeb = 0
        aux = year - 1700 - afterFeb
        # dayOfWeek for 1700/1/1 = 5, Friday
        dayOfWeek  = 5
        # partial sum of days betweem current date and 1700/1/1
        dayOfWeek += (aux + afterFeb) * 365
        # leap year correction
        dayOfWeek += aux / 4 - aux / 100 + (aux + 100) / 400
        # sum monthly and day offsets
        dayOfWeek += offset[month - 1] + (day - 1)
        dayOfWeek %= 7
        return week[int(dayOfWeek)]

    def exCmd(cFile,passedIn):
        try:
            cf = open(cFile)
            if passedIn.find("'") > -1:
                if passedIn.find('"') > -1:
                    if passedIn[0] == '"' and passedIn[-1] == '"':
                        passedIn = passedIn[1:-1]
                        exec('passedIn = "'+passedIn+'"\n'+cf.read())
                    else:
                        print("Invalid argument: "+passedIn)
                else:
                    exec('passedIn = "'+passedIn+'"\n'+cf.read())
            else:
                exec("passedIn = '"+passedIn+"'\n"+cf.read())
            cf.close()

        except Exception as err:
            print("*ERROR* Exception:",str(err),"in",cFile)

        return


    def chkPath(tstPath):
        validPath = True

        simpPath = ""
        if tstPath == []:
            validPath = True
            simpPath = ""
        else:

            savDir = os.getcwd()

            for path in tstPath:
                if path == "":
                    os.chdir(slh)

                elif os.getcwd() == slh and path == "..":
                    validPath = False
                    break

                elif path == ".":
                    continue

                elif path == ".." and len(os.getcwd().split(slh)) == 2:
                    os.chdir(slh)

                elif path == "..":
                    os.chdir("..")

                elif path in os.listdir() and (os.stat(path)[0] & (2**15) == 0):
                    os.chdir(path)

                else:
                    validPath = False
                    simpPath = ""
                    break

            if validPath:
                simpPath = os.getcwd()
            os.chdir(savDir)

        return((validPath,simpPath))

    def absolutePath(argPath,currDir):

        if argPath[0] == slh:
            fullPath = argPath
        elif currDir == slh:
            fullPath = slh+argPath
        else:
            fullPath = currDir+slh+argPath

        if len(fullPath) > 1 and fullPath[-1] == slh:
            fullPath = fullPath[:-1]

        return(fullPath)

    def prDir(dirPath,switches):
        wideCount = 0
        wideCols = int(int(envVars["_scrWidth"])/16)

        def dirLine(fType,dFile,fSize,fTime,swWide):

            if swWide:
                if fType == "D":
                    print("["+dFile[:13]+"]"+" "*(14-len(dFile[:13])),end="")
                else:
                    print(dFile[:15]+" "*(16-len(dFile[:15])),end="")
            else:
                if fType == "D":
                    scrAdj1 = 52 - min(int(envVars["_scrWidth"]),52)
                    scrAdj2 = min(13,65-min(int(envVars["_scrWidth"]),65))
                    print(dFile+" "*(24-len(dFile)-scrAdj1)+"<DIR>"+" "*(18-scrAdj2)+"%2.2i-%2.2i-%4.4i %2.2i:%2.2i" % (fTime[0], fTime[1], fTime[2], fTime[3], fTime[4]))
                else:
                    scrAdj1 = 65 - min(int(envVars["_scrWidth"]),65)
                    print(dFile+" "*(35-len(dFile)+10-len(fSize)-scrAdj1),fSize,"%2.2i-%2.2i-%4.4i %2.2i:%2.2i" % (fTime[0], fTime[1], fTime[2], fTime[3], fTime[4]))

            return()

        def dirSummary(screenlines,swPause,swWide,nLines,nFiles,tFSize,nDirs,availDisk):
            retLines = 2
            if swWide:
                print()
            if swPause and nLines == screenlines:
                anyKey()
                retLines = -2

            scrAdj1 = 65 - min(int(envVars["_scrWidth"]),65)
            print(" "*(4-len(str(nFiles))),nFiles,"File(s)"+" "*(32-len(str(tFSize))-scrAdj1),tFSize,"Bytes.")
            if swPause and nLines+1 == screenlines:
                anyKey()
                retLines = -1
            print(" "*(4-len(str(nDirs))),nDirs,"Dir(s)"+" "*(33-len(str(availDisk))-scrAdj1),availDisk,"Bytes free.")
            return(retLines)


        def txtFileTime(fPath):
            retTime = localtime(max(min(2145916800,os.stat(fPath)[9]),946684800))
            fTime = []
            fTime.append(retTime[1])
            fTime.append(retTime[2])
            fTime.append(retTime[0])
            fTime.append(retTime[3])
            fTime.append(retTime[4])
            return(fTime)

        swError = False
        swPause = False
        swWide = False
        for i in range(len(switches)):
            if switches[i][0] == 'P':
                swPause = True
            elif switches[i][0] == 'W':
                swWide = True
            else:
                print("Illeagal switch:",switches[i][0],"Command Format: DIR/p/w/o:[[-]n,e,s,d]/s [path][file]")
                swError = True
                break
        if swError:
            return

        savDir = os.getcwd()
        fullPath = absolutePath(dirPath,savDir)

        pathDirs = fullPath.split(slh)
        lastDir = pathDirs.pop(-1)

        (validPath, tmpDir) = chkPath(pathDirs)

        if validPath:

            os.chdir(tmpDir)

            if tmpDir == slh:
                pathDirs = [""]
            else:
                pathDirs = tmpDir.split(slh)

            # Check for relative directory from possible mount point root
            if len(pathDirs) == 2:
                if lastDir == ".":
                    os.chdir(slh)
                    lastDir = tmpDir[1:]
                elif lastDir == "..":
                    os.chdir(slh)
                    lastDir = ""

            tmpDir = os.getcwd()

            if lastDir in os.listdir() or lastDir in ".." or "*" in lastDir or "?" in lastDir:

                if "*" in lastDir or "?" in lastDir:
                    nDirs = 0
                    nFiles = 0
                    nLines = 1
                    dPath = tmpDir+(slh if tmpDir[-1] != slh else "")
                    print("Directory of", tmpDir)
                    for _dir in sorted(os.listdir(), key=lambda v: (v.upper(), v[0].isupper())):
                        if os.stat(dPath+_dir)[0] & (2**15) == 0 and _match(lastDir,_dir[:wildcardLen]):
                            fTime = txtFileTime(dPath+_dir)

                            if swWide:
                                if wideCount == wideCols:
                                    wideCount = 0
                                    print()
                                    nLines += 1
                                wideCount += 1

                            if swPause and nLines == int(envVars["_scrHeight"])-1:
                                key = anyKey()
                                nLines = 0
                                if key in "QqCc" and key != "":
                                    break

                            dirLine("D",_dir,0,fTime,swWide)
                            nDirs += 1
                            if not swWide:
                                nLines += 1

                    tFSize = 0
                    try:
                        availDisk = os.statvfs(tmpDir)[1]*os.statvfs(tmpDir)[4]
                    except:
                        availDisk = 0
                    for _dir in sorted(os.listdir(), key=lambda v: (v.upper(), v[0].isupper())):
                        if os.stat(dPath+_dir)[0] & (2**15) != 0 and _match(lastDir,_dir[:wildcardLen]):
                            fSize = str(os.stat(dPath+_dir)[6])
                            tFSize += os.stat(dPath+_dir)[6]
                            fTime = txtFileTime(dPath+_dir)
                            if swWide:
                                if wideCount == wideCols:
                                    wideCount = 0
                                    print()
                                    nLines += 1
                                wideCount += 1
                            if swPause and nLines == int(envVars["_scrHeight"])-1:
                                key = anyKey()
                                nLines = 0
                                if key in "QqCc" and key != "":
                                    break

                            dirLine("F",_dir,fSize,fTime,swWide)
                            nFiles += 1
                            if not swWide:
                                nLines += 1

                    rLines = dirSummary(int(envVars["_scrHeight"])-1,swPause,swWide,nLines,nFiles,tFSize,nDirs,availDisk)
                    if rLines < 0:
                        nLines = -rLines
                    else:
                        nLines += rLines

                elif os.stat((slh if lastDir == "" else lastDir))[0] & (2**15) == 0:

                    nDirs = 2
                    nFiles = 0
                    nLines = 3
                    dPath = tmpDir+(slh if tmpDir[-1] != slh else "")+lastDir
                    # when dPath = "/." sd mount point not included in directory list
                    if dPath == slh+".":
                        dPath = slh
                        lastDir = ""
                    print("Directory of", dPath)
                    if swWide:
                        if wideCols > 1:
                            print("[.]             [..]            ",end="")
                            wideCount = 2
                        else:
                            print("[.]")
                            print("[..]",end="")
                            wideCount = 1
                    else:
                        scrAdj1 = 52 - min(int(envVars["_scrWidth"]),52)
                        print("."+" "*(23-scrAdj1)+"<DIR>")
                        print(".."+" "*(22-scrAdj1)+"<DIR>")
                    for _dir in sorted(os.listdir((slh if lastDir == "" else lastDir)), key=lambda v: (v.upper(), v[0].isupper())):
                        if os.stat(lastDir+slh+_dir)[0] & (2**15) == 0:
                            fTime = txtFileTime(lastDir+slh+_dir)
                            if swWide:
                                if wideCount == wideCols:
                                    wideCount = 0
                                    print()
                                    nLines += 1
                                wideCount += 1
                            if swPause and nLines == int(envVars["_scrHeight"])-1:
                                key = anyKey()
                                nLines = 0
                                if key in "QqCc" and key != "":
                                    break

                            dirLine("D",_dir,0,fTime,swWide)
                            nDirs += 1
                            if not swWide:
                                nLines += 1

                    tFSize = 0
                    try:
                        availDisk = os.statvfs(dPath)[1]*os.statvfs(dPath)[4]
                    except:
                        availDisk = 0
                    for _dir in sorted(os.listdir((slh if lastDir == "" else lastDir)), key=lambda v: (v.upper(), v[0].isupper())):
                        if os.stat(lastDir+slh+_dir)[0] & (2**15) != 0:
                            fSize = str(os.stat(lastDir+slh+_dir)[6])
                            tFSize += int(fSize)
                            fTime = txtFileTime(lastDir+slh+_dir)
                            if swWide:
                                if wideCount == wideCols:
                                    wideCount = 0
                                    print()
                                    nLines += 1
                                wideCount += 1
                            if swPause and nLines == int(envVars["_scrHeight"])-1:
                                key = anyKey()
                                nLines = 0
                                if key in "QqCc" and key != "":
                                    break

                            dirLine("F",_dir,fSize,fTime,swWide)
                            nFiles += 1
                            if not swWide:
                                nLines += 1

                    rLines = dirSummary(int(envVars["_scrHeight"])-1,swPause,swWide,nLines,nFiles,tFSize,nDirs,availDisk)
                    if rLines < 0:
                        nLines = -rLines
                    else:
                        nLines += rLines
                else:
                    nDirs = 0
                    nFiles = 1
                    nLines = 1
                    try:
                        availDisk = os.statvfs(tmpDir)[1]*os.statvfs(tmpDir)[4]
                    except:
                        availDisk = 0
                    print("Directory of",tmpDir)
                    fSize = str(os.stat(lastDir)[6])
                    tFSize = int(fSize)
                    fTime = txtFileTime(lastDir)
                    dirLine("F",lastDir,fSize,fTime,swWide)
                    rLines = dirSummary(int(envVars["_scrHeight"])-1,swPause,swWide,nLines,nFiles,tFSize,nDirs,availDisk)
                    if rLines < 0:
                        nLines = -rLines
                    else:
                        nLines += rLines
            else:
                print("File",dirPath,"not found. (1)")

            os.chdir(savDir)
        else:
            print("File",dirPath,"not found. (2)")

        return

    def filecpy(file1,file2):
        gc.collect()
        fOrig = open(file1)
        fCopy = open(file2, "wb")
        line = fOrig.readline()
        while line != "":
            fCopy.write(bytearray(line))
            line = fOrig.readline()
        fOrig.close()
        fCopy.close()
        gc.collect()
        return

    def setCondCmd(args,i,condResult):
        condCmd = ""
        foundElse = False

        for _ in args[i:]:

            if condResult:
                if _.upper() == "ELSE":
                    break
                condCmd += (_+" ")
            else:
                if foundElse:
                    condCmd += (_+" ")
                else:
                    if _.upper() == "ELSE":
                        foundElse = True

        return condCmd

    def readBATFile(BATfile):
        batIndex = [0]
        batLabels = {}
        batLineNo = 0
        for batLine in BATfile:
            batIndex.append(batIndex[batLineNo]+len(batLine))
            batLineNo += 1
            if batLine.strip() != "":
                if batLine.strip()[0] == ":" and len(batLine.strip().split(" ")[0]) > 1:
                    batLabels[batLine.strip().split(" ")[0][1:]] = [batLineNo,batIndex[batLineNo]]
        BATfile.seek(0)

        return (batIndex,batLabels)

    batEcho = True
    cmd = ""
    condCmd = ""
    os.chdir(slh)
    if "autoexec.bat" in ",".join(os.listdir()).lower().split(','):
        activeBAT = True
        BATfile = open(os.listdir()[(",".join(os.listdir()).lower().split(",")).index("autoexec.bat")])
        (batIndex,batLabels) = readBATFile(BATfile)
        batLineNo = 0
        batParams = []
    else:
        activeBAT = False
        gc.collect()

    while True:
        if condCmd != "":
            cmdLine = condCmd
            condCmd = ""
        else:
            if activeBAT:
                cmdLine = BATfile.readline()
                i=1
                for param in batParams:
                    cmdLine = cmdLine.replace('%'+str(i),param)
                    i+=1
                    if i>9:
                        break

                batLineNo += 1
                if cmdLine == "":
                    activeBAT = False
                    batEcho = True
                    BATfile.close()
                    gc.collect()
                elif batEcho and cmdLine[0] !="@":
                    print(cmdLine,end="")
                elif cmdLine[0] == "@":
                    cmdLine = cmdLine[1:]

            if not activeBAT:
                prompt = "\n"
                for prmpToken in envVars.get('PROMPT','$M$P$G').split("$")[1:]:
                    if prmpToken.upper() == 'M':
                        if 'mem_free' in dir(gc):
                            prompt += "("+str(gc.mem_free())+")"
                    if prmpToken.upper() == 'P':
                        prompt += os.getcwd()
                    if prmpToken.upper() == 'G':
                        prompt += ">"
                cmdLine = input(prompt)

        cmdLine = cmdLine.strip()

        envFound = False
        fndVar = ""
        newCmdLine = ""
        for _ in cmdLine:
            if not envFound:
                if _ == "%":
                    envFound = True
                    doublepct = True
                else:
                    newCmdLine += _
            else:
                if _ == "%":
                    if doublepct:
                        newCmdLine += "%"
                    envFound = False
                    newCmdLine += str(envVars.get(fndVar,""))
                    fndVar = ""
                else:
                    doublepct = False
                    fndVar += _
        cmdLine = newCmdLine

        args = cmdLine.split(" ")

        quotedArg = False
        if len(args) > 1:
            i = 0
            iEnd = len(args)
            for _ in range(0, iEnd):
                if args[i].strip() == "":
                    args.pop(i)
                elif quotedArg:
                    if args[i].find('"') > -1 and args[i].find('"') != len(args[i])-1:
                        break
                    elif args[i][-1] == '"':
                        args[i-1] = args[i-1] + " " + args.pop(i)[:-1]
                        quotedArg = False
                    else:
                        args[i-1] = args[i-1] + " " + args.pop(i)
                elif args[i][0] == '"' and args[i][-1] != '"':
                    args[i] = args[i][1:]
                    i += 1
                    quotedArg = True
                else:
                    i += 1

        if cmdLine != "" and cmdLine[0] != '/':
            switches = (args[0].upper()).split('/')
            cmd = switches.pop(0)
        else:
            switches = ""
            cmd = args[0]

        if quotedArg:
            print("Mismatched quotes.")
            cmd = ""

        if cmd == "" or cmd == "REM":
            continue
        elif cmd == "DIR":
# Command switches /p/w/a:[d]/o:[[-]n,e,s,d]/s needs to be implemented

            if len(args) == 1:

                prDir(os.getcwd(),switches)

            elif len(args) == 2:

                prDir(args[1],switches)

            else:
                print("Too many arguments. Command Format: DIR/p/w/o:[[-]n,e,s,d]/s [path][file]")

        elif cmd == "DATE":
            print("The current date is: "+weekDay()+" %2.2i/%2.2i/%4.4i" % (localtime()[1], localtime()[2], localtime()[0]))

        elif cmd == "TIME":
            print("The current time is: %2.2i:%2.2i:%2.2i" % (localtime()[3], localtime()[4], localtime()[5]))

        elif cmd == "MEM":
            gc.collect()
            #gc.threshold(gc.mem_free() // 4 + gc.mem_alloc())

            print("\n%10i Kb free conventional memory" % (int(gc.mem_free()/1000)))
            print("%10i Kb used conventional memory" % (int(gc.mem_alloc()/1000)))
            if implementation.name.upper() == "MICROPYTHON":
                print("%10i Kb current threshold value" % (int(gc.threshold()/1000)))

                for i in range(len(switches)):
                    if switches[i][0] == 'D':
                        print(mem_info(1))
                    else:
                        print("Illeagal switch:",switches[i][0],"Command Format: mem[/d]")
                        break

        elif cmd == "VER":
            print("PyDOS [Version 1.09]")

        elif cmd == "ECHO":
            if len(args) == 1:
                print("Echo is "+("on." if batEcho else "off."))
            else:
                if args[1].upper() == 'ON':
                    batEcho = True
                elif args[1].upper() == 'OFF':
                    batEcho = False
                else:
                    print(cmdLine[5:])

        elif cmd == "PAUSE":
            anyKey()

        elif cmd[0] == ":" and activeBAT:
            if len(args[0]) <= 1 or len(args) != 1:
                print("Invalid batch label")
                condCmd = "exit"

        elif cmd == "GOTO" and activeBAT:
            if len(args) == 2:
                try:
                    BATfile.seek(batLabels[args[1]][1])
                except:
                    print("Invalid Goto label:",args[1])
                    condCmd = "exit"

                batLineNo = batLabels.get(args[1],[batLineNo+1,0])[0]
            else:
                print("Invalid Goto label:",cmdLine)
                condCmd = "exit"

        elif cmd == "IF" and activeBAT:
            condResult = False

            if len(args) < 3:
                print("Invalid command format:",cmdLine)
                condCmd = "exit"
            else:
                i = 1
                notlogic = False
                if args[1].upper() == "NOT":
                   notlogic = True
                   i = 2

                if args[i].strip().upper() == 'ERRORLEVEL':
                    i += 1
                    if len(args) > i and args[i].isdigit():
                        if str(envVars.get('errorlevel')).isdigit() and int(envVars.get('errorlevel')) == int(args[i]):
                            condResult = True

                        if notlogic:
                            condResult = not condResult

                        i += 1
                        condCmd = setCondCmd(args,i,condResult)
                    else:
                        print("Invalid conditional ERRORLEVEL:",cmdLine)
                        condCmd = "exit"

                elif args[i].strip().upper() == 'EXIST':
                    i += 1
                    if len(args) > i:
                        savDir = os.getcwd()
                        args[i] = absolutePath(args[i],savDir)

                        aPath = args[i].split(slh)
                        newdir = aPath.pop(-1)
                        (validPath, tmpDir) = chkPath(aPath)
                        if tmpDir == "" or tmpDir[-1] != slh:
                            tmpDir += slh

                        if validPath and newdir in os.listdir(tmpDir[:(-1 if tmpDir != slh else None)]):
                            condResult = True
                        if notlogic:
                            condResult = not condResult

                        i += 1
                        condCmd = setCondCmd(args,i,condResult)

                    else:
                        print("Invalid conditional EXIST:",cmdLine)
                        condCmd = "exit"
                else:
                    # string comparison
                    if len(args) > i:
                        string1 = args[i]
                        if "==" in args[i]:
                            string1 = args[i].split("==")[0]
                            if args[i][-1] != "=":
                                string2 = args[i].split("==")[-1]
                            else:
                                i += 1
                                if len(args) > i:
                                    string2 = args[i]
                                else:
                                    print("Invalid string conditional:",cmdLine)
                                    condCmd = "exit"
                        elif len(args) > i+1:
                            i += 1
                            if "==" in args[i]:
                                if len(args[i]) > 2:
                                    string2 = args[i].split("==")[-1]
                                else:
                                    i+=1
                                    if len(args) > i:
                                        string2 = args[i]
                                    else:
                                        print("Invalid string conditional:",cmdLine)
                                        condCmd = "exit"
                            else:
                                print("Invalid string conditional:",cmdLine)
                                condCmd = "exit"
                        else:
                            print("Invalid string conditional:",cmdLine)
                            condCmd = "exit"


                        if condCmd != "exit":
                            if string1 == string2:
                                condResult = True
                            if notlogic:
                                condResult = not condResult

                            i += 1
                            condCmd = setCondCmd(args,i,condResult)
                    else:
                        print("Invalid string conditional:",cmdLine)
                        condCmd = "exit"

        elif cmd == "SET":
            if len(args) == 1:
                for _ in sorted(envVars):
                    print(_+"=",envVars[_],sep="")
            else:
                args = cmdLine.split(" ")
                args.pop(0)
                envCmd = (" ".join(args)).split("=")
                envCmdVar = envCmd.pop(0).strip()

                if len(switches) <= 1:
                    if len(switches) == 0:
                        tmp = "=".join(envCmd).strip()
                    elif switches[0] == 'A':
                        # Replace all possible environment variables with their values
                        envCmd = "=".join(envCmd)
                        for _ in " %*()-+/":
                            envCmd = envCmd.replace(_," ")
                        envCmd = envCmd.split(" ")

                        for _ in envCmd:
                            if _ != "":
                                if _[0].isalpha() or _[0] == "_":
                                    cmdLine = cmdLine.replace(_.strip(),str(envVars.get(_,0)))

                        # Evaluate right sight of = after value substituion
                        args = cmdLine.split(" ")
                        args.pop(0)
                        envCmd = (" ".join(args)).split("=")
                        envCmdVar = envCmd.pop(0).strip()
                        #print("=".join(envCmd))
                        try:
                            envVars[envCmdVar] = str(eval("=".join(envCmd).strip()))
                        except:
                            envVars[envCmdVar] = "0"
                    elif switches[0] == "P":
                        tmp = input("=".join(envCmd).strip()+" ")
                    else:
                        print("Illeagal switch:",'/'.join(switches),"Command Format: SET[/a|/p] [variable = [string|expression]]")

                    if len(switches) == 0 or switches[0] == "P":
                        if tmp != "":
                            envVars[envCmdVar] = tmp
                        else:
                            if envCmdVar == "_scrHeight":
                                envVars["_scrHeight"] = 24
                            elif envCmdVar == "_scrWidth":
                                envVars["_scrWidth"] = 80
                            elif envVars.get(envCmdVar) != None:
                                envVars.pop(envCmdVar)
                else:
                    print("Illeagal switch:",'/'.join(switches),"Command Format: SET[/a|/p] [variable = [string|expression]]")



        elif cmd == "RENAME" or cmd == "REN" or cmd == "MOVE" or cmd == "MV":
# Move command should work like copy where source can be file and target can be a directory
# Wildcard renames should be implemented

            if len(args) == 3:
                aPath = args[1].split(slh)
                newdir = aPath.pop(-1)

# Check that first argument has a valid path and exists
                if chkPath(aPath)[0] and newdir in os.listdir(args[1][0:max(0,len(args[1])-len(newdir)-1)]) and args[1][-1] != slh:

                    aPath2 = args[2].split(slh)
                    newdir2 = aPath2.pop(-1)

# Second argument has valid path
                    if chkPath(aPath2)[0] and args[2][-1] != slh:
# Check that second argument doesn't specify an existing target
                        if newdir2 not in os.listdir(args[2][0:max(0,len(args[2])-len(newdir2)-1)]):
                            currDRen = False
                            savDir = os.getcwd()
                            os.chdir(args[1][0:max(0,len(args[1])-len(newdir)-1)])
                            if os.stat(newdir)[0] & (2**15) == 0:
                                os.chdir(newdir)
                                if savDir == os.getcwd():
                                    currDRen = True
                            os.chdir(savDir)
                            os.rename(args[1],args[2])
                            if currDRen:
                                os.chdir("..")

                        else:
                            print("Target file exists")
                    else:
                        print("Invalid target:",args[2])
                else:
                    print("No such file:",args[1])
            else:
                print("Wrong number of arguments")

        elif cmd == "DELETE" or cmd == "DEL":

            if len(args) == 2:
                savDir = os.getcwd()
                args[1] = absolutePath(args[1],savDir)

                aPath = args[1].split(slh)
                newdir = aPath.pop(-1)
                (validPath, tmpDir) = chkPath(aPath)
                if tmpDir == "" or tmpDir[-1] != slh:
                    tmpDir += slh
                if validPath:
                    if "*" in newdir or "?" in newdir:
                        ans = "Y"
                        if newdir == "*" or newdir == "*.*":
                            ans = input(tmpDir+newdir+", Are you sure (y/n)? ").upper()

                        if ans == "Y":
                            for _dir in os.listdir(tmpDir[:(-1 if tmpDir != slh else None)]):
                                if os.stat(tmpDir+_dir)[0] & (2**15) != 0 and _match(newdir,_dir[:wildcardLen]):
                                    if _dir == _dir[:wildcardLen]:
                                        os.remove(tmpDir+_dir)
                                        print(tmpDir+_dir,"deleted.")
                                    else:
                                        print("Unable to delete: "+tmpDir+_dir+". Filename too long for wildcard operation.")
                    else:
                        if newdir in os.listdir(tmpDir[:(-1 if tmpDir != slh else None)]) and os.stat(tmpDir+newdir)[0] & (2**15) != 0:
                            os.remove(tmpDir+newdir)
                            print(tmpDir+newdir,"deleted.")
                        else:
                            print("Unable to delete: "+tmpDir+newdir+". File not found.")
                else:
                    print("Unable to delete: "+tmpDir+newdir+". File not found.")
            else:
                print("Illeagal Path.")

        elif cmd == "TYPE" or cmd == "MORE":

            if len(args) == 2:

                savDir = os.getcwd()
                args[1] = absolutePath(args[1],savDir)

                aPath = args[1].split(slh)
                newdir = aPath.pop(-1)
                (validPath, tmpDir) = chkPath(aPath)
                if tmpDir == "" or tmpDir[-1] != slh:
                    tmpDir += slh

                if validPath and newdir in os.listdir(tmpDir[:(-1 if tmpDir != slh else None)]) and os.stat(tmpDir+newdir)[0] & (2**15) != 0:
                    swError = False
                    if cmd == "MORE":
                        swPause = True
                    else:
                        swPause = False
                    for i in range(len(switches)):
                        if switches[i][0] == 'P':
                            swPause = True
                        else:
                            print("Illeagal switch:",switches[i][0],"Command Format: TYPE[/p] [path][file]")
                            swError = True
                            break

                    if not swError:
                        key = " "
                        f = open(tmpDir+newdir)
                        line = f.readline()
                        nLines = 0
                        while line != "":
                            istrt = 0
                            while istrt+int(envVars["_scrWidth"]) < len(line):
                                if nLines >= int(envVars["_scrHeight"])-1 and swPause:
                                    key = anyKey()
                                    nLines = 0
                                print(line[istrt:istrt+int(envVars["_scrWidth"])])
                                nLines += 1
                                istrt += int(envVars["_scrWidth"])
                            if key in "QqCc" and key != "":
                                break

                            if nLines >= int(envVars["_scrHeight"])-1 and swPause:
                                key = anyKey()
                                if key in "QqCc" and key != "":
                                    break
                                nLines = 0
                            print(line[istrt:len(line)],end="")
                            line = f.readline()
                            nLines += 1
                        f.close()
                else:
                    print("Unable to display: "+tmpDir+newdir+". File not found.")
            else:
                print("Illeagal Path.")


        elif cmd == "CD":

            if len(args) == 1:
                print(os.getcwd())

            else:
                pathDirs = args[1].split(slh)
                if chkPath(pathDirs)[0]:
                    newdir = os.getcwd()
                    os.chdir(args[1])
                    # if change to .. did nothing assume we're in root of mount point
                    if args[1] == ".." and newdir == os.getcwd():
                        os.chdir(slh)
                else:
                    print("Unable to change to:",args[1])

        elif cmd == "MKDIR":
            if len(args) == 1:
                print("Unable to make .")
            elif len(args) > 2:
                print("Too many arguments")
            else:
                savDir = os.getcwd()
                args[1] = absolutePath(args[1],savDir)

                aPath = args[1].split(slh)
                newdir = aPath.pop(-1)
                (validPath, tmpDir) = chkPath(aPath)
                if tmpDir == "" or tmpDir[-1] != slh:
                    tmpDir += slh

                if validPath:
                    if newdir not in os.listdir(tmpDir[:(-1 if tmpDir != slh else None)]):
                        os.mkdir(tmpDir+newdir)
                    else:
                        print("Target name already exists")
                else:
                    print("Invalid path")

        elif cmd == "RMDIR":
            if len(args) == 1:
                print("The syntax of the command is incorrect")
            elif len(args) > 2:
                print("Too many arguments")
            else:
                savDir = os.getcwd()
                args[1] = absolutePath(args[1],savDir)

                aPath = args[1].split(slh)
                (validPath, tmpDir) = chkPath(aPath)
                if validPath:
# Directory must be empty to be removed
                    if os.listdir(tmpDir) == []:
                        if tmpDir != slh:
                            os.rmdir(tmpDir)
                            if savDir == tmpDir:
                                os.chdir("..")
                        else:
                            print("Can not remove root directory")
                    else:
                        print("The directory is not empty")
                else:
                    print("Invalid path")

        elif cmd == "COPY":

            swError = False
            swConf = False
            for i in range(len(switches)):
                if switches[i][0] == 'Y':
                    swConf = True
                else:
                    print("Illeagal switch:",switches[i][0],"Command Format: COPY[/y] [path]file [path][file]")
                    swError = True
                    break


            if (len(args) == 3 or len(args) == 2) and not swError:
                if len(args) == 2:
                    args.append('.')

                nFiles = 0
                earlyError = False
                trailingSlash = False
                if args[1][-1] == slh:
                    print("The source file cannot be a directory")
                    earlyError = True
                if args[2][-1] == slh:
                    trailingSlash = True
                enteredArg = args[2]

                savDir = os.getcwd()
                args[1] = absolutePath(args[1],savDir)
                args[2] = absolutePath(args[2],savDir)

                aPath = args[1].split(slh)
                newdir = aPath.pop(-1)

# Check that first argument has a valid path, exists and is not a directory file
                (validPath, tmpDir) = chkPath(aPath)
                if not earlyError and validPath and ("*" in newdir or "?" in newdir or (newdir in os.listdir(tmpDir) and os.stat(tmpDir+(slh if tmpDir[-1] != slh else "")+newdir)[0] & (2**15))) != 0:
                    sourcePath = tmpDir
                    if "*" in newdir or "?" in newdir:
                        wildCardOp = True
                    else:
                        wildCardOp = False

                    aPath2 = args[2].split(slh)
                    newdir2 = aPath2.pop(-1)

# Second argument has valid path
                    (validPath, tmpDir) = chkPath(aPath2)
                    if validPath:
                        gc.collect()
                        targetPath = tmpDir

                        if newdir2 in "..":
                            (validPath,tmpDir) = chkPath((targetPath+(slh if targetPath[-1] != slh else "")+newdir2).split(slh))
                            aPath2 = tmpDir.split(slh)
                            newdir2 = aPath2.pop(-1)
                            targetPath = chkPath(aPath2)[1]


# Second argument specifies an existing target
                        if newdir2 in os.listdir(targetPath) or (targetPath == slh and newdir2 == ""):

# Second argument target is a file
                            if os.stat(targetPath+(slh if targetPath[-1] != slh else "")+newdir2)[0] & (2**15) != 0:
                                if trailingSlash:
                                    print("Cannot find the specified path: ",enteredArg)
                                elif wildCardOp:
                                    print("Target must be directory for wildcard copy ",enteredArg)

                                elif sourcePath == targetPath and newdir == newdir2:
                                    print("The file cannot be copied onto itself")

                                elif swConf or input("Overwrite "+args[2]+"? (y/n): ").upper() == "Y":
                                    os.remove(targetPath+(slh if targetPath[-1] != slh else "")+newdir2)
                                    filecpy(sourcePath+(slh if sourcePath[-1] != slh else "")+newdir,targetPath+(slh if targetPath[-1] != slh else "")+newdir2)
                                    nFiles += 1

# Second argument target is a directory
                            else:
                                if sourcePath == "" or sourcePath[-1] != slh:
                                    sourcePath += slh
                                if targetPath == "" or targetPath[-1] != slh:
                                    targetPath += slh

                                if wildCardOp:
                                    ans = ""
                                    for _dir in os.listdir(sourcePath[:(-1 if sourcePath != slh else None)]):
                                        if os.stat(sourcePath+_dir)[0] & (2**15) != 0 and _match(newdir,_dir[:wildcardLen]):
                                            print("copy",sourcePath+_dir,"to",targetPath+newdir2+("" if newdir2 == "" else slh)+_dir)
                                            if _dir in os.listdir(targetPath+newdir2):
                                                if sourcePath == targetPath+newdir2+("" if newdir2 == "" else slh):
                                                    print("The file cannot be copied onto itself")
                                                    break
                                                else:
                                                    if ans != "A" and not swConf:
                                                        ans = input("Overwrite "+targetPath+newdir2+("" if newdir2 == "" else slh)+_dir+"? (y/n/(q)uit/(a)ll): ").upper()
                                                    if ans  == "Y" or ans == "A" or swConf:
                                                        filecpy(sourcePath+_dir,targetPath+newdir2+("" if newdir2 == "" else slh)+_dir)
                                                        nFiles += 1
                                                    elif ans == "Q":
                                                        break
                                            else:
                                                filecpy(sourcePath+_dir,targetPath+newdir2+("" if newdir2 == "" else slh)+_dir)
                                                nFiles += 1
                                            gc.collect()

                                elif newdir in os.listdir(targetPath+newdir2):
                                    if sourcePath == targetPath+newdir2+("" if newdir2 == "" else slh):
                                        print("The file cannot be copied onto itself")
                                    elif swConf or input("Overwrite "+targetPath+newdir2+("" if newdir2 == "" else slh)+newdir+"? (y/n): ").upper() == "Y":
                                        os.remove(targetPath+newdir2+("" if newdir2 == "" else slh)+newdir)
                                        filecpy(sourcePath+newdir,targetPath+newdir2+("" if newdir2 == "" else slh)+newdir)
                                        nFiles += 1

                                else:
                                    filecpy(sourcePath+newdir,targetPath+newdir2+("" if newdir2 == "" else slh)+newdir)
                                    nFiles += 1
                        elif not validPath:
                            print("Invalid Target:",args[2])
# Second argument is a new file
                        else:
                            if trailingSlash or wildCardOp:
                                if wildCardOp:
                                    print("Target must be directory for wildcard copy ",enteredArg)
                                else:
                                    print("Cannot find the specified path: ",enteredArg)
                            else:
                                filecpy(sourcePath+(slh if sourcePath[-1] != slh else "")+newdir,targetPath+(slh if targetPath[-1] != slh else "")+newdir2)
                                nFiles += 1

                        print(" "*7,nFiles,"files(s) copied.")
                    else:
                        print("Invalid destination:",args[2])
                else:
                    print("No such file:",args[1])
            else:
                if not swError:
                    print("Wrong number of arguments")

        elif cmd == "EXIT":
            if activeBAT:
                if len(args) > 1:
                    condCmd = "set errorlevel="+args[1]
                activeBAT = False
                batEcho = True
                BATfile.close()
            else:
                threadFound = True
                try:
                    testthreadLock = threadLock
                except:
                    threadFound = False

                if threadFound:
                    import _thread

                    if threadLock.locked():
                        threadLock.release()
                    else:
                        threadLock.acquire()

                break

        else:
            savDir = os.getcwd()
            args[0] = absolutePath(args[0],savDir)

            aPath = args[0].split(slh)
            newdir = aPath.pop(-1)
            (validPath, tmpDir) = chkPath(aPath)
            if tmpDir == "" or tmpDir[-1] != slh:
                tmpDir += slh


            if len(args) == 1:
                passedIn = ""
                batParams = []
            elif len(args) > 1:
                passedIn = args[1]
                batParams = args[1:]

            gc.collect()
            batFound = ""
            curDLst = os.listdir(tmpDir[:(-1 if tmpDir != slh else None)])
            curDLst_LC = ",".join(curDLst).lower().split(",")
            if  ((newdir.split("."))[-1]).upper() == "PY":
                if validPath and newdir in curDLst and os.stat(tmpDir+newdir)[0] & (2**15) != 0:

                    exCmd(tmpDir+newdir,passedIn)
                else:
                    print("Illegal command:",args[0])
            elif ((newdir.split("."))[-1]).upper() == "BAT":
                if validPath and newdir in curDLst and os.stat(tmpDir+newdir)[0] & (2**15) != 0:
                    #batFound = 0
                    batFound = newdir
                else:
                    print("Illegal command:",args[0])
            elif validPath:
                if newdir.lower()+".py" in curDLst_LC:
                    curDLst_indx = curDLst_LC.index(newdir.lower()+".py")
                    if os.stat(tmpDir+curDLst[curDLst_indx])[0] & (2**15) != 0:
                        exCmd(tmpDir+curDLst[curDLst_indx],passedIn)
                elif newdir.lower()+".bat" in curDLst_LC:
                    curDLst_indx = curDLst_LC.index(newdir.lower()+".bat")
                    if os.stat(tmpDir+curDLst[curDLst_indx])[0] & (2**15) != 0:
                        batFound = curDLst[curDLst_indx]
                else:
                    print("Illegal command:",cmdLine.split(" ")[0])
            else:
                print("Illegal command:",cmdLine.split(" ")[0])

            if batFound != "":
                if activeBAT:
                    BATfile.close()
                BATfile = open(tmpDir+batFound)
                activeBAT = True
                batEcho = True
                (batIndex,batLabels) = readBATFile(BATfile)
                batLineNo = 0

            gc.collect()

PyDOS()



import network, socket

config = {}
envfound = True
try:
    envfile = open('/.env')
except:
    envfound = False

if envfound:
    for line in envfile:
        try:
            config[line.split('=')[0].strip()] = line.split('=')[1].strip()
        except:
            pass
    envfile.close()

# Check .env has updated credentials
if not envfound or config.get('CIRCUITPY_WIFI_SSID','') == '':
    assert False, ("/.env has not been updated with your unique keys and data")

# AP info
SSID=config['CIRCUITPY_WIFI_SSID'] # Network SSID
KEY=config['CIRCUITPY_WIFI_PASSWORD']  # Network key

PORT = 80
HOST = "www.google.com"

# Init wlan module and connect to network
print("Trying to connect. Note this may take a while...")

wlan = network.WLAN(network.STA_IF)
wlan.active(True)
wlan.connect(SSID, KEY)

# We should have a valid IP now via DHCP
print("Wi-Fi Connected ", wlan.ifconfig())

# Get addr info via DNS
addr = socket.getaddrinfo(HOST, PORT)[0][4]
print(addr)

# Create a new socket and connect to addr
client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
client.connect(addr)

# Set timeout
client.settimeout(3.0)

# Send HTTP request and recv response
client.send("GET / HTTP/1.1\r\nHost: %s\r\n\r\n"%(HOST))
print((client.recv(1024)).decode())
# Close socket
client.close()


HOST = 'wifitest.adafruit.com'
# Get addr info via DNS
addr = socket.getaddrinfo(HOST, PORT)[0][4]
print(addr)

# Create a new socket and connect to addr
client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
client.connect(addr)

# Set timeout
client.settimeout(3.0)

# Send HTTP request and recv response
client.send("GET /testwifi/index.html HTTP/1.1\r\nHost: %s\r\n\r\n"%(HOST))
print("-" * 40)
print((client.recv(1024)).decode())
print("-" * 40)


# Close socket
client.close()



from bbq10keyboard import BBQ10Keyboard, STATE_PRESS, STATE_RELEASE, STATE_LONG_PRESS
from adafruit_stmpe610 import Adafruit_STMPE610_SPI
import adafruit_ili9341
import adafruit_sdcard
import digitalio
import displayio
import neopixel
import storage
import board
import time
import os

# Optional Qwiic test
try:
    import adafruit_pct2075
except Exception as e:
    print('Skipping Qwiic test,', e)

all_passed = True

# Release any resources currently in use for the displays
displayio.release_displays()

spi = board.SPI()
tft_cs = board.D9
tft_dc = board.D10
touch_cs = board.D6
sd_cs = board.D5
neopix_pin = board.D11

display_bus = displayio.FourWire(spi, command=tft_dc, chip_select=tft_cs)
display = adafruit_ili9341.ILI9341(display_bus, width=320, height=240)

print('Display: Pass? (you tell me)')

# Touch Screen
print('Touch the screen')
try:
    touch = Adafruit_STMPE610_SPI(spi, digitalio.DigitalInOut(touch_cs))
    while touch.buffer_empty:
        pass

    print('Touch: Pass,', touch.read_data())
except Exception as e:
    print('Touch: Fail,', e)
    all_passed = False

# SD Card
try:
    sdcard = adafruit_sdcard.SDCard(spi, digitalio.DigitalInOut(sd_cs))
    vfs = storage.VfsFat(sdcard)
    storage.mount(vfs, '/sd')

    # Expect at least one file on the card
    if len(os.listdir('/sd/')) > 0:
        print('SD-Card: Pass')
    else:
        raise Exception('No files')
except Exception as e:
    print('SD-Card: Fail,', e)
    all_passed = False

# Neopixel test, pink!
try:
    pixels = neopixel.NeoPixel(neopix_pin, 1)
    pixels[0] = 0xFF00FF
    time.sleep(0.3)
    pixels.brightness = 0.1
    print('Neopixel: Pass')
except Exception as e:
    print('Neopixel: Fail,', e)
    all_passed = False

try:
    i2c = board.I2C()
except Exception as e:
    print('I2C: Fail,', e)
    all_passed = False

# Optional
try:
    pct = adafruit_pct2075.PCT2075(i2c)
    print('Temperature: %.2f C' % pct.temperature)
except Exception as e:
    print('Qwiic: Fail (might be fine),', e)

# Keyboard test helper
def await_key(kbd, key, name=''):
    if name == '':
        name = key

    print('Press %s' % name)

    while kbd.key_count < 2:
        pass

    keys = kbd.keys
    res = (keys[0] == (1, key) and keys[1] == (3, key))
    if not res:
        raise Exception('Expected %s but got %s' % (name, keys))

# Keyboard
try:
    kbd = BBQ10Keyboard(i2c)

    # Test a key from each column
    await_key(kbd, 'q')
    await_key(kbd, 's')
    await_key(kbd, 'p')
    await_key(kbd, 'b')
    await_key(kbd, 'm')
    await_key(kbd, 'f')

    # All the extra keys
    await_key(kbd, '\x06', 'L1')
    await_key(kbd, '\x11', 'L2')
    await_key(kbd, '\x07', 'R1')
    await_key(kbd, '\x12', 'R2')
    await_key(kbd, '\x01', 'UP')
    await_key(kbd, '\x02', 'DOWN')
    await_key(kbd, '\x03', 'LEFT')
    await_key(kbd, '\x04', 'RIGHT')

    print('Keyboard: Pass')
except Exception as e:
    print('Keyboard: Fail', e)
    all_passed = False

if all_passed:
    print('All tests passed!')
else:
    print('Some fails!')



#! /usr/bin/python

# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from basictoken import BASICToken as Token
from flowsignal import FlowSignal
from sys import implementation
from time import sleep
import math
import random
try:
    from pydos_ui import input
except:
    pass
try:
    from pydos_hw import Pydos_hw
    sndPin = Pydos_hw.sndPin
except:
    sndPin = None

if implementation.name.upper() == 'MICROPYTHON':
    if sndPin:
        from machine import PWM
    from time import ticks_ms as monotonic

elif implementation.name.upper() == 'CIRCUITPYTHON':
    from time import monotonic
    if sndPin:
        try: # temporary? until broadcom port supports pwmio
            from pwmio import PWMOut
        except:
            pass
else:
    import winsound
    from time import monotonic

import gc
gc.collect()

"""Implements a BASIC array, which may have up
to three dimensions of fixed size.

"""
class BASICArray:

    def __init__(self, dimensions):
        """Initialises the object with the specified
        number of dimensions. Maximum number of
        dimensions is three

        :param dimensions: List of array dimensions and their
        corresponding sizes

        """
        self.dims = min(3,len(dimensions))

        if self.dims == 0:
            raise SyntaxError("Zero dimensional array specified")

        # Check for invalid sizes and ensure int
        for i in range(self.dims):
            if dimensions[i] < 0:
                raise SyntaxError("Negative array size specified")
            # Allow sizes like 1.0f, but not 1.1f
            if int(dimensions[i]) != dimensions[i]:
                raise SyntaxError("Fractional array size specified")
            dimensions[i] = int(dimensions[i])

        if self.dims == 1:
            self.data = [None for x in range(dimensions[0])]
        elif self.dims == 2:
            self.data = [[None for x in range(dimensions[1])] for x in range(dimensions[0])]
        else:
            self.data = [[[None for x in range(dimensions[2])] for x in range(dimensions[1])] for x in range(dimensions[0])]

#    def pretty_print(self):
#        print(str(self.data))


"""Implements a BASIC parser that parses a single
statement when supplied.

"""
class BASICParser:

    def __init__(self):
        # Symbol table to hold variable names mapped
        # to values
        self.__symbol_table = {}

        # Stack on which to store operands
        # when evaluating expressions
        self.__operand_stack = []

        # List to hold contents of DATA statement
        self.__data_values = []

        # These values will be
        # initialised on a per
        # statement basis
        self.__tokenlist = []
        self.__tokenindex = None

        # used to determine when to initalize extant loop variables
        self.last_flowsignal = None

        # Set to keep track of print column across multiple print statements
        self.__prnt_column = 0

        #file handle list
        self.__file_handles = {}

        if implementation.name.upper() == 'MICROPYTHON':
            if sndPin:
                try:
                    self.__pwm = PWM(sndPin,freq=0)
                except:
                    self.__pwm = PWM(sndPin)
            if 'duty_u16' in dir(self.__pwm):
                self.__pwm.duty_u16(0)
            else:
                self.__pwm.duty(0)


    def parse(self, tokenlist, line_number, cstmt_number, infile, tmpfile, datastmts):
        """Must be initialised with the list of
        BTokens to be processed. These tokens
        represent a BASIC statement without
        its corresponding line number.

        :param tokenlist: The tokenized program statement
        :param line_number: The line number of the statement
        :param cstmt_number: Which statement in a multistatment line

        :return: The FlowSignal to indicate to the program
        how to branch if necessary, None otherwise

        """

        self.__tokenlist = tokenlist
        self.__tokenindex = 0


        # This block locates the index number of the first token in the statement
        # referenced by che cstmt_number argument
        indx = 0
        colon_count = 0
        if cstmt_number > 0:
            for e in tokenlist:
                if e.category == Token.COLON:
                    colon_count += 1
                    self.__tokenindex = indx + 1
                indx += 1
                if colon_count >= cstmt_number:
                    break

        # Remember the line number to aid error reporting
        self.__line_number = line_number

        # Assign the first token
        self.__token = self.__tokenlist[self.__tokenindex]

        flow = self.__stmt(infile,tmpfile,datastmts)

        # If the statement returns an EXECUTE flowsignal then it's a conditional
        # and we need to parse the THEN or ELSE block depending on where the
        # __ifstmt method left the current __tokenindex
        if flow and (flow.ftype == FlowSignal.EXECUTE):

            # Find the number of compound statements in the current
            # conditional block (to an ELSE or end of line)
            number_of_stmts = 1
            for e in tokenlist[self.__tokenindex:]:
                if e.category == Token.COLON:
                    number_of_stmts += 1
                elif e.category == Token.ELSE:
                    break

            recur_tokenindex = self.__tokenindex
            self.__tokenindex = 0
            for cstmtNo in range(0,number_of_stmts):
                try:
                #if True:
                    tmp_flow = self.parse(tokenlist[recur_tokenindex:],line_number,cstmtNo,infile,tmpfile,datastmts)

                except RuntimeError as err:
                    raise RuntimeError(str(err))

                if tmp_flow:
                    break

            return tmp_flow

        else:
            return flow


    def __advance(self):
        """Advances to the next token

        """
        # Move to the next token
        self.__tokenindex += 1

        # Acquire the next token if there any left
        if not self.__tokenindex >= len(self.__tokenlist):
            self.__token = self.__tokenlist[self.__tokenindex]

    def __consume(self, expected_category):
        """Consumes a token from the list

        """
        if self.__token.category == expected_category:
            self.__advance()

        else:
            raise RuntimeError('Expecting ' + Token.catnames[expected_category] +
                               ' in line ' + str(self.__line_number))

    def __stmt(self,infile,tmpfile,datastmts):
        """Parses a program statement

        :return: The FlowSignal to indicate to the program
        how to branch if necessary, None otherwise

        """

        if self.__token.category in [Token.FOR, Token.IF, Token.NEXT,
                                     Token.ON]:
            return self.__compoundstmt()

        else:
            return self.__simplestmt(infile,tmpfile,datastmts)

    def __simplestmt(self,infile,tmpfile,datastmts):
        """Parses a non-compound program statement

        :return: The FlowSignal to indicate to the program
        how to branch if necessary, None otherwise

        """
        if self.__token.category == Token.NAME:
            self.__assignmentstmt()
            return None

        elif self.__token.category == Token.PRINT:
            self.__printstmt()
            return None

        elif self.__token.category == Token.LET:
            self.__letstmt()
            return None

        elif self.__token.category == Token.GOTO:
            return self.__gotostmt()

        elif self.__token.category == Token.GOSUB:
            return self.__gosubstmt()

        elif self.__token.category == Token.RETURN:
            return self.__returnstmt()

        elif self.__token.category == Token.STOP:
            return self.__stopstmt()

        elif self.__token.category == Token.INPUT:
            self.__inputstmt()
            return None

        elif self.__token.category == Token.DIM:
            self.__dimstmt()
            return None

        elif self.__token.category == Token.RANDOMIZE:
            self.__randomizestmt()
            return None

        elif self.__token.category == Token.DATA:
            self.__datastmt()
            return None

        elif self.__token.category == Token.READ:
            self.__readstmt(infile,tmpfile,datastmts)
            return None

        elif self.__token.category == Token.OPEN:
            return self.__openstmt()

        elif self.__token.category == Token.CLOSE:
            self.__closestmt()
            return None

        elif self.__token.category == Token.FSEEK:
            self.__fseekstmt()
            return None

        elif self.__token.category == Token.RESTORE:
            self.__restorestmt(datastmts)
            return None

        elif self.__token.category == Token.SOUND:
            self.__soundstmt()
            return None

        else:
            # Ignore comments, but raise an error
            # for anything else
            if self.__token.category != Token.REM:
                raise RuntimeError('Expecting program statement in line '
                                   + str(self.__line_number))

    def __printstmt(self):
        """Parses a PRINT statement, causing
        the value that is on top of the
        operand stack to be printed on
        the screen.

        """
        self.__advance()   # Advance past PRINT token

        fileIO = False
        if self.__token.category == Token.HASH:
            fileIO = True

            # Process the # keyword
            self.__consume(Token.HASH)

            # Acquire the file number
            self.__expr()
            filenum = self.__operand_stack.pop()

            if self.__file_handles.get(filenum) == None:
                raise RuntimeError("PRINT: file #"+str(filenum)+" not opened in line " + str(self.__line_number))

            # Process the comma
            if self.__tokenindex < len(self.__tokenlist) and self.__token.category != Token.COLON:
                self.__consume(Token.COMMA)

        # Check there are items to print
        last_token_cat = None
        if not self.__tokenindex >= len(self.__tokenlist) and self.__token.category != Token.COLON:
            last_token_cat = self.__token.category
            prntTab = (self.__token.category == Token.TAB)
            self.__logexpr()

            #if type(self.__operand_stack[-1]) == tuple and self.__operand_stack[-1][0] == "TAB":
            if prntTab:
                if self.__prnt_column >= len(self.__operand_stack[-1]):
                    if fileIO:
                        self.__file_handles[filenum].write("\n")
                    else:
                        print()
                    self.__prnt_column = 0

                current_pr_column = len(self.__operand_stack[-1]) - self.__prnt_column
                self.__prnt_column = len(self.__operand_stack.pop()) - 1
                if current_pr_column > 1:
                    if fileIO:
                        self.__file_handles[filenum].write(" "*(current_pr_column-1))
                    else:
                        print(" "*(current_pr_column-1), end="")
            else:
                self.__prnt_column += len(str(self.__operand_stack[-1]))
                if fileIO:
                    self.__file_handles[filenum].write('%s' %(self.__operand_stack.pop()))
                else:
                    print(self.__operand_stack.pop(), end='')

            while self.__token.category == Token.COMMA or self.__token.category == Token.SEMICOLON:
                last_token_cat = self.__token.category
                self.__advance()
                prntTab = (self.__token.category == Token.TAB)
                if not self.__tokenindex >= len(self.__tokenlist) and self.__token.category != Token.COLON:
                    last_token_cat = None
                    self.__logexpr()

                    #if type(self.__operand_stack[-1]) == tuple and self.__operand_stack[-1][0] == "TAB":
                    if prntTab:
                        if self.__prnt_column >= len(self.__operand_stack[-1]):
                            if fileIO:
                                self.__file_handles[filenum].write("\n")
                            else:
                                print()
                            self.__prnt_column = 0
                        current_pr_column = len(self.__operand_stack[-1]) - self.__prnt_column
                        if fileIO:
                            self.__file_handles[filenum].write(" "*(current_pr_column-1))
                        else:
                            print(" "*(current_pr_column-1), end="")
                        self.__prnt_column = len(self.__operand_stack.pop()) - 1
                    else:
                        self.__prnt_column += len(str(self.__operand_stack[-1]))
                        if fileIO:
                            self.__file_handles[filenum].write('%s' %(self.__operand_stack.pop()))
                        else:
                            print(self.__operand_stack.pop(), end='')

                else:
                    break

        # Final newline
        if last_token_cat != Token.SEMICOLON:
            if fileIO:
                self.__file_handles[filenum].write("\n")
            else:
                print()
            self.__prnt_column = 0

    def __letstmt(self):
        """Parses a LET statement,
        consuming the LET keyword.
        """
        self.__advance()  # Advance past the LET token
        self.__assignmentstmt()

    def __gotostmt(self):
        """Parses a GOTO statement

        :return: A FlowSignal containing the target line number
        of the GOTO

        """
        self.__advance()  # Advance past GOTO token
        self.__expr()

        # Set up and return the flow signal
        return FlowSignal(ftarget=self.__operand_stack.pop())

    def __gosubstmt(self):
        """Parses a GOSUB statement

        :return: A FlowSignal containing the first line number
        of the subroutine

        """

        self.__advance()  # Advance past GOSUB token
        self.__expr()

        # Set up and return the flow signal
        return FlowSignal(ftarget=self.__operand_stack.pop(),
                          ftype=FlowSignal.GOSUB)

    def __returnstmt(self):
        """Parses a RETURN statement"""

        self.__advance()  # Advance past RETURN token

        # Set up and return the flow signal
        return FlowSignal(ftype=FlowSignal.RETURN)

    def __stopstmt(self):
        """Parses a STOP statement"""

        self.__advance()  # Advance past STOP token

        for handles in self.__file_handles:
            self.__file_handles[handles].close()
        self.__file_handles.clear()

        return FlowSignal(ftype=FlowSignal.STOP)

    def __assignmentstmt(self):
        """Parses an assignment statement,
        placing the corresponding
        variable and its value in the symbol
        table.

        """
        left = self.__token.lexeme  # Save lexeme of
                                    # the current token
        self.__advance()

        if self.__token.category == Token.LEFTPAREN:
            # We are assiging to an array
            self.__arrayassignmentstmt(left)

        else:
            # We are assigning to a simple variable
            self.__consume(Token.ASSIGNOP)
            self.__logexpr()

            # Check that we are using the right variable name format
            right = self.__operand_stack.pop()

            if left.endswith('$') and not isinstance(right, str):
                raise SyntaxError('Syntax error: Attempt to assign non string to string variable' +
                                  ' in line ' + str(self.__line_number))

            elif not left.endswith('$') and isinstance(right, str):
                raise SyntaxError('Syntax error: Attempt to assign string to numeric variable' +
                                  ' in line ' + str(self.__line_number))

            self.__symbol_table[left] = right

    def __dimstmt(self):
        """Parses  DIM statement and creates a symbol
        table entry for an array of the specified
        dimensions.

        """
        self.__advance()  # Advance past DIM keyword

        # Extract the array name, append a suffix so
        # that we can distinguish from simple variables
        # in the symbol table
        name = self.__token.lexeme + '_array'
        self.__advance()  # Advance past array name

        self.__consume(Token.LEFTPAREN)

        # Extract the dimensions
        dimensions = []
        if not self.__tokenindex >= len(self.__tokenlist):
            self.__expr()
            dimensions.append(self.__operand_stack.pop())

            while self.__token.category == Token.COMMA:
                self.__advance()  # Advance past comma
                self.__expr()
                dimensions.append(self.__operand_stack.pop())

        self.__consume(Token.RIGHTPAREN)

        if len(dimensions) > 3:
            raise SyntaxError("Maximum number of array dimensions is three " +
                              "in line " + str(self.__line_number))

        self.__symbol_table[name] = BASICArray(dimensions)

    def __arrayassignmentstmt(self, name):
        """Parses an assignment to an array variable

        :param name: Array name

        """
        self.__consume(Token.LEFTPAREN)

        # Capture the index variables
        # Extract the dimensions
        indexvars = []
        if not self.__tokenindex >= len(self.__tokenlist):
            self.__expr()
            indexvars.append(self.__operand_stack.pop())

            while self.__token.category == Token.COMMA:
                self.__advance()  # Advance past comma
                self.__expr()
                indexvars.append(self.__operand_stack.pop())

        try:
            BASICarray = self.__symbol_table[name + '_array']

        except KeyError:
            raise KeyError('Array - ' + name + ' could not be found in line ' +
                           str(self.__line_number))

        if BASICarray.dims != len(indexvars):
            raise IndexError('Incorrect number of indices applied to array ' +
                             'in line ' + str(self.__line_number))

        self.__consume(Token.RIGHTPAREN)
        self.__consume(Token.ASSIGNOP)

        self.__logexpr()

        # Check that we are using the right variable name format
        right = self.__operand_stack.pop()

        if name.endswith('$') and not isinstance(right, str):
            raise SyntaxError('Attempt to assign non string to string array' +
                              ' in line ' + str(self.__line_number))

        elif not name.endswith('$') and isinstance(right, str):
            raise SyntaxError('Attempt to assign string to numeric array' +
                              ' in line ' + str(self.__line_number))

        # Assign to the specified array index
        try:
            if len(indexvars) == 1:
                BASICarray.data[indexvars[0]-1] = right

            elif len(indexvars) == 2:
                BASICarray.data[indexvars[0]-1][indexvars[1]-1] = right

            elif len(indexvars) == 3:
                BASICarray.data[indexvars[0]-1][indexvars[1]-1][indexvars[2]-1] = right

        except IndexError:
            raise IndexError('Array index out of range in line ' +
                             str(self.__line_number))

    def __openstmt(self):
        """Parses an open statement, opens the indicated file and
        places the file handle into handle table

        """
        self.__advance() # Advance past OPEN token

        # Acquire the filename
        self.__logexpr()
        filename = self.__operand_stack.pop()

        # Process the FOR keyword
        self.__consume(Token.FOR)

        if self.__token.lexeme == "INPUT":
            accessMode = "r"
        elif self.__token.lexeme == "APPEND":
            accessMode = "r+"
        elif self.__token.lexeme == "OUTPUT":
            accessMode = "w+"
        else:
            raise SyntaxError('Invalid Open access mode in line ' + str(self.__line_number))

        self.__advance() # Advance past acess type

        if self.__token.lexeme != "AS":
            raise SyntaxError('Expecting AS in line ' + str(self.__line_number))

        self.__advance() # Advance past AS keyword

        #if self.__token.category != Token.HASH:
            #raise SyntaxError('Expecting (#filenum) in line ' + str(self.__line_number))

        #self.__advance() # Advance past Hashmark (#)

        # Process the # keyword
        self.__consume(Token.HASH)

        # Acquire the file number
        self.__expr()
        filenum = self.__operand_stack.pop()

        branchOnError = False
        if self.__token.category == Token.ELSE:
            branchOnError = True
            self.__advance() # Advance past ELSE

            if self.__token.category == Token.GOTO:
                self.__advance()    # Advance past optional GOTO

            self.__expr()

        if self.__file_handles.get(filenum) != None:
            if branchOnError:
                return FlowSignal(ftarget=self.__operand_stack.pop())
            else:
                raise RuntimeError("File #",filenum," already opened in line " + str(self.__line_number))

        try:
            self.__file_handles[filenum] = open(filename,accessMode)

        except:
            if branchOnError:
                return FlowSignal(ftarget=self.__operand_stack.pop())
            else:
                raise RuntimeError('File '+filename+' could not be opened in line ' + str(self.__line_number))

        if accessMode == "r+":
            self.__file_handles[filenum].seek(0)
            filelen = 0
            for lines in self.__file_handles[filenum]:
                filelen += (len(lines)+(0 if implementation.name.upper() in ['MICROPYTHON','CIRCUITPYTHON'] else 1))

            self.__file_handles[filenum].seek(filelen)

        return None

    def __closestmt(self):
        """Parses a close, closes the file and removes
        the file handle from the handle table

        """

        self.__advance() # Advance past CLOSE token

        # Process the # keyword
        self.__consume(Token.HASH)

        # Acquire the file number
        self.__expr()
        filenum = self.__operand_stack.pop()

        if self.__file_handles.get(filenum) == None:
            raise RuntimeError("CLOSE: file #"+str(filenum)+" not opened in line " + str(self.__line_number))

        self.__file_handles[filenum].close()
        self.__file_handles.pop(filenum)

    def __fseekstmt(self):
        """Parses an fseek statement, seeks the indicated file position

        """

        self.__advance() # Advance past FSEEK token

        # Process the # keyword
        self.__consume(Token.HASH)

        # Acquire the file number
        self.__expr()
        filenum = self.__operand_stack.pop()

        if self.__file_handles.get(filenum) == None:
            raise RuntimeError("FSEEK: file #"+str(filenum)+" not opened in line " + str(self.__line_number))

        # Process the comma
        self.__consume(Token.COMMA)

        # Acquire the file position
        self.__expr()

        self.__file_handles[filenum].seek(self.__operand_stack.pop())

    def __inputstmt(self):
        """Parses an input statement, extracts the input
        from the user and places the values into the
        symbol table

        """
        self.__advance()  # Advance past INPUT token

        fileIO = False
        if self.__token.category == Token.HASH:
            fileIO = True

            # Process the # keyword
            self.__consume(Token.HASH)

            # Acquire the file number
            self.__expr()
            filenum = self.__operand_stack.pop()

            if self.__file_handles.get(filenum) == None:
                raise RuntimeError("INPUT: file #"+str(filenum)+" not opened in line " + str(self.__line_number))

            # Process the comma
            self.__consume(Token.COMMA)

        prompt = '? '
        if self.__token.category == Token.STRING:
            if fileIO:
                raise SyntaxError('Input prompt specified for file I/O ' +
                                'in line ' + str(self.__line_number))

            # Acquire the input prompt
            self.__logexpr()
            prompt = self.__operand_stack.pop()
            self.__consume(Token.COMMA)

        # Acquire the comma separated input variables
        variables = []
        if not self.__tokenindex >= len(self.__tokenlist):
            if self.__token.category != Token.NAME:
                raise ValueError('Expecting NAME in INPUT statement ' +
                                 'in line ' + str(self.__line_number))
            variables.append(self.__token.lexeme)
            self.__advance()  # Advance past variable

            while self.__token.category == Token.COMMA:
                self.__advance()  # Advance past comma
                variables.append(self.__token.lexeme)
                self.__advance()  # Advance past variable

        # Gather input into the variables
        if fileIO:
            #inputvals = self.__file_handles[filenum].readline()[:(-2 if implementation.name.upper() ==
                        #'MICROPYTHON' else -1)].split(',', (len(variables)-1))
            inputvals = ((self.__file_handles[filenum].readline().replace("\n","")).replace("\r","")).split(',', (len(variables)-1))
        else:
            # kfw inputvals = self.input_keyboard(prompt).split(',', (len(variables)-1))
            inputvals = input(prompt).split(',', (len(variables)-1))

        for variable in variables:
            left = variable

            try:
                right = inputvals.pop(0)

                if left.endswith('$'):
                    self.__symbol_table[left] = str(right)

                elif not left.endswith('$'):
                    try:
                        self.__symbol_table[left] = int(right)

                    except ValueError:
                        raise ValueError('String input provided to a numeric variable ' +
                                         'in line ' + str(self.__line_number))

            except IndexError:
                # No more input to process
                pass

    def __restorestmt(self,datastmts):

        self.__advance() # Advance past RESTORE token

        # Acquire the line number
        self.__expr()

        self.__data_values.clear()
        datastmts.restore(self.__operand_stack.pop())

    def __datastmt(self):
        """Parses a DATA statement"""

    def __readstmt(self,infile,tmpfile,datastmts):
        """Parses a READ statement."""

        self.__advance()  # Advance past READ token

        # Acquire the comma separated input variables
        variables = []
        if not self.__tokenindex >= len(self.__tokenlist):
            variables.append(self.__token.lexeme)
            self.__advance()  # Advance past variable

            while self.__token.category == Token.COMMA:
                self.__advance()  # Advance past comma
                variables.append(self.__token.lexeme)
                self.__advance()  # Advance past variable


        # Check that we have enough data values to fill the
        # variables
        #if len(variables) > len(self.__data_values):
            #raise RuntimeError('Insufficient constants supplied to READ ' +
                               #'in line ' + str(self.__line_number))


        # Gather input from the DATA statement into the variables
        for variable in variables:

            if len(self.__data_values) < 1:
                self.__data_values = datastmts.readData(self.__line_number,infile,tmpfile)

            left = variable
            #right = readlist.pop(0)
            right = self.__data_values.pop(0)

            if left.endswith('$'):
                # Python inserts quotes around input data
                if isinstance(right, int):
                    raise ValueError('Non-string input provided to a string variable ' +
                                     'in line ' + str(self.__line_number))

                else:
                    self.__symbol_table[left] = right

            elif not left.endswith('$'):
                try:
                    #self.__symbol_table[left] = int(right)
                    self.__symbol_table[left] = right

                except ValueError:
                    raise ValueError('String input provided to a numeric variable ' +
                                     'in line ' + str(self.__line_number))

    def __soundstmt(self):
        """Parses a SOUND statement"""

        self.__advance()  # Advance past SOUND token

        # Acquire the comma separated values
        self.__expr()
        freq = self.__operand_stack.pop()

        self.__consume(Token.COMMA)
        self.__expr()
        duration = self.__operand_stack.pop()

        if self.__token.category == Token.COMMA:
            self.__advance()
            self.__expr()
            volume = self.__operand_stack.pop()
        else:
            volume = 800

        if implementation.name.upper() == 'MICROPYTHON':
            if sndPin:
                self.__pwm.freq(freq)
                if "duty_u16" in dir(self.__pwm):
                    self.__pwm.duty_u16(volume)
                    sleep(duration/18.2)
                    self.__pwm.duty_u16(0)
                else:
                    self.__pwm.duty(int((volume/65535)*1023))
                    sleep(duration/18.2)
                    self.__pwm.duty(0)
        elif implementation.name.upper() == 'CIRCUITPYTHON':
            try:
                Pydos_hw.sndGPIO.deinit() # Workaround for ESP32-S2 GPIO issue
                audioPin = PWMOut(sndPin, duty_cycle=volume, frequency=freq, variable_frequency=True)
                sleep(duration/18.2)
                audioPin.deinit()
                Pydos_hw.quietSnd() # Workaround for ESP32-S2 GPIO issue
            except:
                pass
        else:
            winsound.Beep(freq,int(self.__operand_stack.pop()*1000/18.2))


    def __expr(self):
        """Parses a numerical expression consisting
        of two terms being added or subtracted,
        leaving the result on the operand stack.

        """
        self.__term()  # Pushes value of left term
                       # onto top of stack

        while self.__token.category in [Token.PLUS, Token.MINUS]:
            savedcategory = self.__token.category
            self.__advance()
            self.__term()  # Pushes value of right term
                           # onto top of stack
            rightoperand = self.__operand_stack.pop()
            leftoperand = self.__operand_stack.pop()

            if savedcategory == Token.PLUS:
                self.__operand_stack.append(leftoperand + rightoperand)

            else:
                self.__operand_stack.append(leftoperand - rightoperand)

    def __term(self):
        """Parses a numerical expression consisting
        of two factors being multiplied together,
        leaving the result on the operand stack.

        """
        self.__sign = 1  # Initialise sign to keep track of unary
                         # minuses
        self.__factor()  # Leaves value of term on top of stack

        while self.__token.category in [Token.TIMES, Token.DIVIDE, Token.MODULO]:
            savedcategory = self.__token.category
            self.__advance()
            self.__sign = 1  # Initialise sign
            self.__factor()  # Leaves value of term on top of stack
            rightoperand = self.__operand_stack.pop()
            leftoperand = self.__operand_stack.pop()

            if savedcategory == Token.TIMES:
                self.__operand_stack.append(leftoperand * rightoperand)

            elif savedcategory == Token.DIVIDE:
                self.__operand_stack.append(leftoperand / rightoperand)

            else:
                self.__operand_stack.append(leftoperand % rightoperand)

    def __factor(self):
        """Evaluates a numerical expression
        and leaves its value on top of the
        operand stack.

        """
        if self.__token.category == Token.PLUS:
            self.__advance()
            self.__factor()

        elif self.__token.category == Token.MINUS:
            self.__sign = -self.__sign
            self.__advance()
            self.__factor()

        elif self.__token.category == Token.UNSIGNEDINT:
            self.__operand_stack.append(self.__sign*int(self.__token.lexeme))
            self.__advance()

        elif self.__token.category == Token.UNSIGNEDFLOAT:
            self.__operand_stack.append(self.__sign*float(self.__token.lexeme))
            self.__advance()

        elif self.__token.category == Token.STRING:
            self.__operand_stack.append(self.__token.lexeme)
            self.__advance()

        elif self.__token.category == Token.NAME and \
             self.__token.category not in Token.functions:
            # Check if this is a simple or array variable
            if (self.__token.lexeme + '_array') in self.__symbol_table:
                # Capture the current lexeme
                arrayname = self.__token.lexeme + '_array'

                # Array must be processed
                # Capture the index variables
                self.__advance()  # Advance past the array name

                try:
                    self.__consume(Token.LEFTPAREN)
                except RuntimeError:
                    raise RuntimeError('Array used without index in line ' +
                                     str(self.__line_number))

                indexvars = []
                if not self.__tokenindex >= len(self.__tokenlist):
                    self.__expr()
                    indexvars.append(self.__operand_stack.pop())

                    while self.__token.category == Token.COMMA:
                        self.__advance()  # Advance past comma
                        self.__expr()
                        indexvars.append(self.__operand_stack.pop())

                BASICarray = self.__symbol_table[arrayname]
                arrayval = self.__get_array_val(BASICarray, indexvars)

                if arrayval != None:
                    self.__operand_stack.append(self.__sign*arrayval)

                else:
                    raise IndexError('Empty array value returned in line ' +
                                     str(self.__line_number))

            elif self.__token.lexeme in self.__symbol_table:
                # Simple variable must be processed
                self.__operand_stack.append(self.__sign*self.__symbol_table[self.__token.lexeme])

            else:
                raise RuntimeError('Name ' + self.__token.lexeme + ' is not defined' +
                                   ' in line ' + str(self.__line_number))

            self.__advance()

        elif self.__token.category == Token.LEFTPAREN:
            self.__advance()

            # Save sign because expr() calls term() which resets
            # sign to 1
            savesign = self.__sign
            self.__logexpr()  # Value of expr is pushed onto stack

            if savesign == -1:
                # Change sign of expression
                self.__operand_stack[-1] = -self.__operand_stack[-1]

            self.__consume(Token.RIGHTPAREN)

        elif self.__token.category in Token.functions:
            self.__operand_stack.append(self.__evaluate_function(self.__token.category))

        else:
            raise RuntimeError('Expecting factor in numeric expression (' +
                               self.__token.lexeme+ ') in line ' + str(self.__line_number))

    def __get_array_val(self, BASICarray, indexvars):
        """Extracts the value from the given BASICArray at the specified indexes

        :param BASICarray: The BASICArray
        :param indexvars: The list of indexes, one for each dimension

        :return: The value at the indexed position in the array

        """
        if BASICarray.dims != len(indexvars):
            raise IndexError('Incorrect number of indices applied to array ' +
                             'in line ' + str(self.__line_number))

        # Fetch the value from the array
        try:
            if len(indexvars) == 1:
                arrayval = BASICarray.data[indexvars[0]-1]

            elif len(indexvars) == 2:
                arrayval = BASICarray.data[indexvars[0]-1][indexvars[1]-1]

            elif len(indexvars) == 3:
                arrayval = BASICarray.data[indexvars[0]-1][indexvars[1]-1][indexvars[2]-1]

        except IndexError:
            raise IndexError('Array index out of range in line ' +
                             str(self.__line_number))

        return arrayval

    def __compoundstmt(self):
        """Parses compound statements,
        specifically if-then-else and
        loops

        :return: The FlowSignal to indicate to the program
        how to branch if necessary, None otherwise

        """
        if self.__token.category == Token.FOR:
            return self.__forstmt()

        elif self.__token.category == Token.NEXT:
            return self.__nextstmt()

        elif self.__token.category == Token.IF:
            return self.__ifstmt()

        elif self.__token.category == Token.ON:
            return self.__ongosubstmt()

    def __ifstmt(self):
        """Parses if-then-else
        statements

        :return: The FlowSignal to indicate to the program
        how to branch if necessary, None otherwise

        """

        self.__advance()  # Advance past IF token
        self.__logexpr()

        # Save result of expression
        saveval = self.__operand_stack.pop()

        # Process the THEN part and save the jump value
        self.__consume(Token.THEN)

        if self.__token.category != Token.UNSIGNEDINT:
            if saveval:
                return FlowSignal(ftype=FlowSignal.EXECUTE)
        else:
            self.__expr()

            # Jump if the expression evaluated to True
            if saveval:
                # Set up and return the flow signal
                return FlowSignal(ftarget=self.__operand_stack.pop())

        # advance to ELSE
        while self.__tokenindex < len(self.__tokenlist) and self.__token.category != Token.ELSE:
            self.__advance()

        # See if there is an ELSE part
        if self.__token.category == Token.ELSE:
            self.__advance()

            if self.__token.category != Token.UNSIGNEDINT:
                return FlowSignal(ftype=FlowSignal.EXECUTE)
            else:
                self.__expr()

                # Set up and return the flow signal
                return FlowSignal(ftarget=self.__operand_stack.pop())

        else:
            # No ELSE action
            return None

    def __forstmt(self):
        """Parses for loops

        :return: The FlowSignal to indicate that
        a loop start has been processed

        """

        # Set up default loop increment value
        step = 1

        self.__advance()  # Advance past FOR token

        # Process the loop variable initialisation
        loop_variable = self.__token.lexeme  # Save lexeme of
                                             # the current token

        if loop_variable.endswith('$'):
            raise SyntaxError('Syntax error: Loop variable is not numeric' +
                              ' in line ' + str(self.__line_number))

        self.__advance()  # Advance past loop variable
        self.__consume(Token.ASSIGNOP)
        self.__expr()

        # Check that we are using the right variable name format
        # for numeric variables
        start_val = self.__operand_stack.pop()

        # Advance past the 'TO' keyword
        self.__consume(Token.TO)

        # Process the terminating value
        self.__expr()
        end_val = self.__operand_stack.pop()

        # Check if there is a STEP value
        increment = True
        if not self.__tokenindex >= len(self.__tokenlist):
            self.__consume(Token.STEP)

            # Acquire the step value
            self.__expr()
            step = self.__operand_stack.pop()

            # Check whether we are decrementing or
            # incrementing
            if step == 0:
                raise IndexError('Zero step value supplied for loop' +
                                 ' in line ' + str(self.__line_number))

            elif step < 0:
                increment = False

        # Now determine the status of the loop

        # Note that we cannot use the presence of the loop variable in
        # the symbol table for this test, as the same variable may already
        # have been instantiated elsewhere in the program

        # Initialize the loop variable anytime the for statement is reached
        # from a statement other than an active NEXT.
        from_next = False
        if self.last_flowsignal:
            if self.last_flowsignal.ftype == FlowSignal.LOOP_REPEAT:
                from_next = True

        if not from_next:
            self.__symbol_table[loop_variable] = start_val

        else:
            # We need to modify the loop variable
            # according to the STEP value
            self.__symbol_table[loop_variable] += step

        # If the loop variable has reached the end value,
        # remove it from the set of extant loop variables to signal that
        # this is the last loop iteration
        stop = False
        if increment and self.__symbol_table[loop_variable] > end_val:
            stop = True

        elif not increment and self.__symbol_table[loop_variable] < end_val:
            stop = True

        if stop:
            # Loop must terminate
            return FlowSignal(ftype=FlowSignal.LOOP_SKIP,
                              ftarget=loop_variable)
        else:
            # Set up and return the flow signal
            return FlowSignal(ftype=FlowSignal.LOOP_BEGIN,floop_var=loop_variable)

    def __nextstmt(self):
        """Processes a NEXT statement that terminates
        a loop

        :return: A FlowSignal indicating that a loop
        has been processed

        """

        self.__advance()  # Advance past NEXT token

        # Process the loop variable initialisation
        loop_variable = self.__token.lexeme  # Save lexeme of
                                             # the current token

        if loop_variable.endswith('$'):
            raise SyntaxError('Syntax error: Loop variable is not numeric' +
                              ' in line ' + str(self.__line_number))

        return FlowSignal(ftype=FlowSignal.LOOP_REPEAT,floop_var=loop_variable)

    def __ongosubstmt(self):
        """Process the ON-GOSUB/GOTO statement

        :return: A FlowSignal indicating the subroutine line number
        if the condition is true, None otherwise

        """

        self.__advance()  # Advance past ON token

        self.__expr()

        # Save result of expression
        saveval = self.__operand_stack.pop()

        if self.__token.category == Token.GOTO:
            self.__consume(Token.GOTO)
            branchtype = 1
        else:
            self.__consume(Token.GOSUB)
            branchtype = 2

        branch_values = []
        # Acquire the comma separated values
        if not self.__tokenindex >= len(self.__tokenlist):
            self.__expr()
            branch_values.append(self.__operand_stack.pop())

            while self.__token.category == Token.COMMA:
                self.__advance()  # Advance past comma
                self.__expr()
                branch_values.append(self.__operand_stack.pop())

        if saveval < 1 or saveval > len(branch_values) or len(branch_values) == 0:
            return None
        elif branchtype == 1:
            return FlowSignal(ftarget=branch_values[saveval-1])
        else:
            return FlowSignal(ftarget=branch_values[saveval-1],
                              ftype=FlowSignal.GOSUB)


    def __relexpr(self):
        """Parses a relational expression
        """
        self.__expr()

        # Since BASIC uses same operator for both
        # assignment and equality, we need to check for this
        if self.__token.category == Token.ASSIGNOP:
            self.__token.category = Token.EQUAL

        if self.__token.category in [Token.LESSER, Token.LESSEQUAL,
                              Token.GREATER, Token.GREATEQUAL,
                              Token.EQUAL, Token.NOTEQUAL]:
            savecat = self.__token.category
            self.__advance()
            self.__expr()

            right = self.__operand_stack.pop()
            left = self.__operand_stack.pop()

            if savecat == Token.EQUAL:
                self.__operand_stack.append(left == right)  # Push True or False

            elif savecat == Token.NOTEQUAL:
                self.__operand_stack.append(left != right)  # Push True or False

            elif savecat == Token.LESSER:
                self.__operand_stack.append(left < right)  # Push True or False

            elif savecat == Token.GREATER:
                self.__operand_stack.append(left > right)  # Push True or False

            elif savecat == Token.LESSEQUAL:
                self.__operand_stack.append(left <= right)  # Push True or False

            elif savecat == Token.GREATEQUAL:
                self.__operand_stack.append(left >= right)  # Push True or False

    def __logexpr(self):
        """Parses a logical expression
        """
        self.__notexpr()

        while self.__token.category in [Token.OR, Token.AND]:
            savecat = self.__token.category
            self.__advance()
            self.__notexpr()

            right = self.__operand_stack.pop()
            left = self.__operand_stack.pop()

            if savecat == Token.OR:
                self.__operand_stack.append(left or right)  # Push True or False

            elif savecat == Token.AND:
                self.__operand_stack.append(left and right)  # Push True or False

    def __notexpr(self):
        """Parses a logical not expression
        """
        if self.__token.category == Token.NOT:
            self.__advance()
            self.__relexpr()
            right = self.__operand_stack.pop()
            self.__operand_stack.append(not right)
        else:
            self.__relexpr()

    def __evaluate_function(self, category):
        """Evaluate the function in the statement
        and return the result.

        :return: The result of the function

        """

        self.__advance()  # Advance past function name

        # Process arguments according to function
        if category == Token.RND:
            return random.random()
        """
        if category == Token.PI:
            return math.pi

        if category == Token.RNDINT:
            self.__consume(Token.LEFTPAREN)

            self.__expr()
            lo = self.__operand_stack.pop()

            self.__consume(Token.COMMA)

            self.__expr()
            hi = self.__operand_stack.pop()

            self.__consume(Token.RIGHTPAREN)

            try:
                return random.randint(lo, hi)

            except ValueError:
                raise ValueError("Invalid value supplied to RNDINT in line " +
                                 str(self.__line_number))
        """
        if category == Token.MAX:
            self.__consume(Token.LEFTPAREN)

            self.__expr()
            value_list = [self.__operand_stack.pop()]

            while self.__token.category == Token.COMMA:
                self.__advance() # Advance past comma
                self.__expr()
                value_list.append(self.__operand_stack.pop())

            self.__consume(Token.RIGHTPAREN)

            try:
                return max(*value_list)

            except TypeError:
                raise TypeError("Invalid type supplied to MAX in line " +
                                 str(self.__line_number))

        if category == Token.MIN:
            self.__consume(Token.LEFTPAREN)

            self.__expr()
            value_list = [self.__operand_stack.pop()]

            while self.__token.category == Token.COMMA:
                self.__advance() # Advance past comma
                self.__expr()
                value_list.append(self.__operand_stack.pop())

            self.__consume(Token.RIGHTPAREN)

            try:
                return min(*value_list)

            except TypeError:
                raise TypeError("Invalid type supplied to MIN in line " +
                                 str(self.__line_number))

        if category == Token.POW:
            self.__consume(Token.LEFTPAREN)

            self.__expr()
            base = self.__operand_stack.pop()

            self.__consume(Token.COMMA)

            self.__expr()
            exponent = self.__operand_stack.pop()

            self.__consume(Token.RIGHTPAREN)

            try:
                return math.pow(base, exponent)

            except ValueError:
                raise ValueError("Invalid value supplied to POW in line " +
                                 str(self.__line_number))
        """
        if category == Token.TERNARY:
            self.__consume(Token.LEFTPAREN)

            self.__logexpr()
            condition = self.__operand_stack.pop()

            self.__consume(Token.COMMA)

            self.__expr()
            whentrue = self.__operand_stack.pop()

            self.__consume(Token.COMMA)

            self.__expr()
            whenfalse = self.__operand_stack.pop()

            self.__consume(Token.RIGHTPAREN)

            return whentrue if condition else whenfalse
        """
        if category == Token.MID:
            self.__consume(Token.LEFTPAREN)

            self.__expr()
            instring = self.__operand_stack.pop()

            self.__consume(Token.COMMA)

            self.__expr()
            start = self.__operand_stack.pop()

            if self.__token.category == Token.COMMA:
                self.__advance() # Advance past comma
                self.__expr()
                end = start + self.__operand_stack.pop() - 1
            else:
                end = None

            start = start - 1

            self.__consume(Token.RIGHTPAREN)

            try:
                return instring[start:end]

            except TypeError:
                raise TypeError("Invalid type supplied to MID$ in line " +
                                 str(self.__line_number))

        if category == Token.INSTR:
            self.__consume(Token.LEFTPAREN)

            self.__expr()
            hackstackstring = self.__operand_stack.pop()
            if not isinstance(hackstackstring, str):
                raise TypeError("Invalid type supplied to INSTR in line " +
                                 str(self.__line_number))

            self.__consume(Token.COMMA)

            self.__expr()
            needlestring = self.__operand_stack.pop()

            start = end = None
            if self.__token.category == Token.COMMA:
                self.__advance() # Advance past comma
                self.__expr()
                start = self.__operand_stack.pop()

                if self.__token.category == Token.COMMA:
                    self.__advance() # Advance past comma
                    self.__expr()
                    end = self.__operand_stack.pop()

            self.__consume(Token.RIGHTPAREN)

            try:
                return hackstackstring.find(needlestring, start, end) + 1

            except TypeError:
                raise TypeError("Invalid type supplied to INSTR in line " +
                                 str(self.__line_number))

        self.__consume(Token.LEFTPAREN)

        self.__expr()
        value = self.__operand_stack.pop()

        self.__consume(Token.RIGHTPAREN)

        if category == Token.SQR:
            try:
                return math.sqrt(value)

            except ValueError:
                raise ValueError("Invalid value supplied to SQR in line " +
                                 str(self.__line_number))

        elif category == Token.ABS:
            try:
                return abs(value)

            except ValueError:
                raise ValueError("Invalid value supplied to ABS in line " +
                                 str(self.__line_number))

        elif category == Token.ATN:
            try:
                return math.atan(value)

            except ValueError:
                raise ValueError("Invalid value supplied to ATN in line " +
                                 str(self.__line_number))

        elif category == Token.COS:
            try:
                return math.cos(value)

            except ValueError:
                raise ValueError("Invalid value supplied to COS in line " +
                                 str(self.__line_number))

        elif category == Token.EXP:
            try:
                return math.exp(value)

            except ValueError:
                raise ValueError("Invalid value supplied to EXP in line " +
                                 str(self.__line_number))

        elif category == Token.INT:
            try:
                return math.floor(value)

            except ValueError:
                raise ValueError("Invalid value supplied to INT in line " +
                                 str(self.__line_number))

        elif category == Token.ROUND:
            try:
                return round(value)

            except TypeError:
                raise TypeError("Invalid type supplied to LEN in line " +
                                 str(self.__line_number))

        elif category == Token.LOG:
            try:
                return math.log(value)

            except ValueError:
                raise ValueError("Invalid value supplied to LOG in line " +
                                 str(self.__line_number))

        elif category == Token.SIN:
            try:
                return math.sin(value)

            except ValueError:
                raise ValueError("Invalid value supplied to SIN in line " +
                                 str(self.__line_number))

        elif category == Token.TAN:
            try:
                return math.tan(value)

            except ValueError:
                raise ValueError("Invalid value supplied to TAN in line " +
                                 str(self.__line_number))

        elif category == Token.CHR:
            try:
                return chr(value)

            except TypeError:
                raise TypeError("Invalid type supplied to CHR$ in line " +
                                 str(self.__line_number))

            except ValueError:
                raise ValueError("Invalid value supplied to CHR$ in line " +
                                 str(self.__line_number))

        elif category == Token.ASC:
            try:
                return ord(value)

            except TypeError:
                raise TypeError("Invalid type supplied to ASC in line " +
                                 str(self.__line_number))

            except ValueError:
                raise ValueError("Invalid value supplied to ASC in line " +
                                 str(self.__line_number))
        elif category == Token.STR:
            return str(value)

        elif category == Token.VAL:
            try:
#                numeric = float(value)
#                if numeric.is_integer():
#                    return int(numeric)
                return float(value)

            # Like other BASIC variants, non-numeric strings return 0
            except ValueError:
                return 0

        elif category == Token.LEN:
            try:
                return len(value)

            except TypeError:
                raise TypeError("Invalid type supplied to LEN in line " +
                                 str(self.__line_number))

        elif category == Token.UPPER:
            if not isinstance(value, str):
                raise TypeError("Invalid type supplied to UPPER$ in line " +
                                 str(self.__line_number))

            return value.upper()

        elif category == Token.LOWER:
            if not isinstance(value, str):
                raise TypeError("Invalid type supplied to LOWER$ in line " +
                                 str(self.__line_number))

            return value.lower()

        elif category == Token.TAB:
            if type(value) == int:
                return " " * value

            else:
                raise TypeError("Invalid type supplied to TAB in line " +
                                 str(self.__line_number))

        else:
            raise SyntaxError("Unrecognised function in line " +
                              str(self.__line_number))

    def __randomizestmt(self):
        """Implements a function to seed the random
        number generator

        """
        self.__advance()  # Advance past RANDOMIZE token

        if not self.__tokenindex >= len(self.__tokenlist):
            self.__expr()  # Process the seed
            seed = self.__operand_stack.pop()
        else:
            seed = int(monotonic())

        random.seed(seed)



import os
from pydos_ui import Pydos_ui
try:
    from pydos_ui import input
except:
    pass
#import uselect

def viewFile(args):

    #def kbdInterrupt():

        #spoll = uselect.poll()
        #spoll.register(sys.stdin,uselect.POLLIN)

        #while not spoll.poll(0):
            #time.sleep(.25)

        #cmnd = sys.stdin.read(1)

        #spoll.unregister(sys.stdin)

        #return(cmnd)

    def chkPath(tstPath):
        validPath = True

        simpPath = ""
        if tstPath == []:
            validPath = True
            simpPath = ""
        else:

            savDir = os.getcwd()

            for path in tstPath:
                if path == "":
                    os.chdir("/")

                elif os.getcwd() == "/" and path == "..":
                    validPath = False
                    break

                elif path == ".":
                    continue

                elif path == ".." and len(os.getcwd().split('/')) == 2:
                    os.chdir('/')

                elif path == "..":
                    os.chdir("..")

                elif path in os.listdir() and (os.stat(path)[0] & (2**15) == 0):
                    os.chdir(path)

                else:
                    validPath = False
                    simpPath = ""
                    break

            if validPath:
                simpPath = os.getcwd()
            os.chdir(savDir)

        return((validPath,simpPath))

    def absolutePath(argPath,currDir):

        if argPath[0] == '/':
            fullPath = argPath
        elif currDir == '/':
            fullPath = '/'+argPath
        else:
            fullPath = currDir+'/'+argPath

        if len(fullPath) > 1 and fullPath[-1] == '/':
            fullPath = fullPath[:-1]

        return(fullPath)


    if __name__ == "PyDOS":
        global envVars

        scrLines = int(envVars["_scrHeight"])
        scrWidth = int(envVars["_scrWidth"])
    else:
        scrLines = 24
        scrWidth = 80

    savDir = os.getcwd()
    args = absolutePath(args,savDir)

    aPath = args.split("/")
    newdir = aPath.pop(-1)
    (validPath, tmpDir) = chkPath(aPath)

    if validPath and newdir in os.listdir(tmpDir) and os.stat(tmpDir+("/" if tmpDir[-1] != "/" else "")+newdir)[0] & (2**15) != 0:
        f = open(args)
        index = [0]
        currLineNum = 0
        maxRead = 0
        eof = -1
        for i in range(scrLines):
            line = f.readline()
            if line != "":
                index.append(index[i]+len(line))
                print()
                print((line[:-1])[:scrWidth],end="")
                currLineNum += 1
                maxRead += 1
            else:
                eof = currLineNum
                break

        cmnd = ""
        seqCnt = 0
        while cmnd.upper() != "Q":
            #cmnd = kbdInterrupt()
            cmnd = Pydos_ui.read_keyboard(1)

            if ord(cmnd) == 27 and seqCnt == 0:
                seqCnt = 1
            elif ord(cmnd) == 91 and seqCnt == 1:
                seqCnt = 2
            elif ord(cmnd) == 65 and seqCnt == 2:
                # Up Arrow
                seqCnt = 0
                if currLineNum > scrLines:
                    currLineNum -= 1
                    print (chr(27)+"[1;1H"+chr(27)+"M",end="")
                    #print (chr(27)+"[2;0H"+chr(27)+"[T",end="")
                    f.seek(index[currLineNum-scrLines])
                    print((f.readline()[:-1])[:scrWidth],end="")

            elif ord(cmnd) == 66 and seqCnt == 2:
                # Down Arrow
                seqCnt = 0
                if eof == -1 or currLineNum < eof:
                    f.seek(index[currLineNum])
                    line = f.readline()
                    if line != "":
                        if currLineNum == maxRead:
                            index.append(index[currLineNum]+len(line))
                            maxRead += 1
                        print(chr(27)+"["+str(scrLines)+";1H"+chr(27)+"D",end="")
                        #print(chr(27)+"["+str(scrLines+1)+";0H"+chr(27)+"[S",end="")
                        print((line[:-1])[:scrWidth],end="")
                        currLineNum += 1
                    else:
                        eof = currLineNum

            elif ord(cmnd) == 67 and seqCnt == 2:
                # Right Arrow
                seqCnt = 0
            elif ord(cmnd) == 68 and seqCnt == 2:
                # Left Arrow
                seqCnt = 0
            else:
                seqCnt = 0


        print(chr(27)+"["+str(scrLines)+";1H",end="")
        f.close()
    else:
        print("Unable to display: "+args+". File not found.")

if __name__ != "PyDOS":
    passedIn = ""

if passedIn == "":
    passedIn = input("Enter filename:")

viewFile(passedIn)



# Developed by Scott Shawcroft (@tannewt) https://github.com/tannewt/basicpython
#
# This is an experiment to edit Python code like BASIC was edited.
# The idea is imagining this as the default mode on a Raspberry Pi 400
#
import os

if __name__ == "PyDOS":
    global envVars
    nCols = int((int(envVars["_scrWidth"])-1)/16)
else:
    nCols = 4

print("# Basic Python v0")

program = []
top = {}
no_ready = False

while True:
    if no_ready:
        line = input()
    else:
        line = input("READY.\n")
    lineno = None
    if " " in line:
        command, remainder = line.split(" ", 1)
        lineno = None
        try:
            lineno = int(command, 10)
        except ValueError:
            pass
    else:
        command = line
        remainder = ""
    no_ready = False
    if lineno is not None:
        if lineno < 1:
            print("Line must be 1+")
        else:
            if lineno >= len(program):
                program.extend([""] * (lineno - len(program)))
            program[lineno - 1] = remainder
            no_ready = True
    elif command.lower() == 'new':
        program.clear()
    elif command.lower() == 'dir':
        wideCount = 0
        for dir in os.listdir():
            if os.stat(dir)[0] & (2**15) == 0:
                if (wideCount % nCols) == 0:
                    print()
                wideCount += 1
                print(("<"+dir+">"+" "*(13-len(dir)))[:15]+" ",end="")

        for dir in os.listdir():
            if os.stat(dir)[0] & (2**15) != 0:
                if (wideCount % nCols) == 0:
                    print()
                wideCount += 1
                print((dir+" "*(15-len(dir)))[:15]+" ",end="")
        print()
    elif command.lower() == "del":
        if remainder.strip(" \"") != "":
            if remainder.strip(" \"") in os.listdir():
                os.remove(remainder.strip(" \""))
            else:
                print("Unable to delete: "+remainder.strip(" \"")+". File not found.")
        else:
            print("No file specified!")
    elif command.lower() == "list":
        for i, line in enumerate(program):
            if line:
                print(i+1, line)
    elif command.lower() == "run":
        isolated = {}
        try:
            exec("\n".join(program), isolated, isolated)
        except Exception as e:
            print(e)
    elif command.lower() == "save":
        filename = remainder.strip(" \"")
        with open(filename, "w") as f:
            f.write("\n".join(program))
    elif command.lower() == "load":
        filename = remainder.strip(" \"")
        with open(filename, "r") as f:
            program = f.readlines()
            program = [line.strip("\r\n") for line in program]
    elif command.lower() in ["exit","quit","dos"]:
        break
    else:
        try:
            exec(line, top, top)
        except Exception as e:
            print(e)



#! /usr/bin/python

# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Class representing a BASIC DATA Statements.
This is a list of DATA statements, ordered by
line number.

"""

from basictoken import BASICToken as Token
from lexer import Lexer

class BASICData:

    def __init__(self):
        # Dictionary to represent data
        # statements, keyed by line number
        self.__data = {}

        # Data pointer
        self.__next_data = 0

    def delete(self):
        self.__data.clear()
        self.__next_data = 0

    def delData(self,line_number):
        if self.__data.get(line_number) != None:
            del self.__data[line_number]

    def addData(self,line_number,fIndex):
        """
        Adds the supplied token list
        to the program's DATA store. The first token should
        be the line number. If a token list with the
        same line number already exists, this is
        replaced.

        line_number: Basic program line number of DATA statement
        fIndex: if >= 0: location in loaded program file of statment
                if < 0: Location in temporary basic workfile

        """

        try:
            self.__data[line_number] = fIndex

        except TypeError as err:
            raise TypeError("Invalid line number: " + str(err))

    def readData(self,read_line_number,infile,tmpfile):

        if len(self.__data) == 0:
            raise RuntimeError('No DATA statements available to READ ' +
                               'in line ' + str(read_line_number))

        data_values = []

        line_numbers = list(self.__data.keys())
        line_numbers.sort()
        #print("+++",self.__next_data,line_numbers)


        if self.__next_data == 0:
            self.__next_data = line_numbers[0]
        elif line_numbers.index(self.__next_data) < len(line_numbers)-1:
            self.__next_data = line_numbers[line_numbers.index(self.__next_data)+1]
        else:
            raise RuntimeError('No DATA statements available to READ ' +
                               'in line ' + str(read_line_number))

        if self.__data[self.__next_data] >= 0:
            infile.seek(self.__data[self.__next_data])
            tokenlist = Lexer().tokenize((infile.readline().replace("\n","")).replace("\r",""))[1:]
        else:
            tmpfile.seek(-(self.__data[self.__next_data]+1))
            tokenlist = Lexer().tokenize((tmpfile.readline().replace("\n","")).replace("\r",""))[1:]

        sign = 1
        for token in tokenlist[1:]:
            if token.category != Token.COMMA:
                if token.category == Token.STRING:
                    data_values.append(token.lexeme)
                elif token.category == Token.UNSIGNEDINT:
                    data_values.append(sign*int(token.lexeme))
                elif token.category == Token.UNSIGNEDFLOAT:
                    data_values.append(sign*eval(token.lexeme))
                elif token.category == Token.MINUS:
                    sign = -1
                #else:
                    #data_values.append(token.lexeme)
            else:
                sign = 1


        return data_values

    def restore(self,restoreLineNo):
        if restoreLineNo == 0 or self.__data.get(restoreLineNo) != None:

            if restoreLineNo == 0:
                self.__next_data = restoreLineNo
            else:

                line_numbers = list(self.__data.keys())
                line_numbers.sort()

                indexln = line_numbers.index(restoreLineNo)

                if indexln == 0:
                    self.__next_data = 0
                else:
                    self.__next_data = line_numbers[indexln-1]
        else:
            raise RuntimeError('Attempt to RESTORE but no DATA ' +
                               'statement at line ' + str(restoreLineNo))


try:
    from pydos_ui import input
except:
    pass
def run(runargv):
    cf = open(runargv)
    exec(cf.read())
    cf.close()

if __name__ != "PyDOS":
    passedIn = ""

if passedIn == "":
    passedIn = input("Enter python file to run: ")

runargv = passedIn
passedIn = ""
run(runargv)



# PyDOS Board Configuration for 'TinyPICO with ESP32-PICO-D4'

Pydos_pins = {
    'sndPin' : (27,"IO27"),
    'I2C_NUM' : (1,None),
    'SCL' : (22,"SCL IO22"),
    'SDA' : (21,"SDA IO21"),
    'SPI_NUM' : (1,None),
    'SCK' : (18,"SCK IO18"),
    'MOSI' : (23,"MO IO23"),
    'MISO' : (19,"MI IO19"),
    'CS' : (5,"SS IO5")
}



import machine, time

mdays = [0,31,29,31,30,31,30,31,31,30,31,30,31]
rtcBase=0x4005c000
atomicBSet=0x2000

print("The current date is: %2.2i/%2.2i/%4.4i" % (time.localtime()[1], time.localtime()[2], time.localtime()[0]))
newDate = input("Enter the new date (mm-dd-yy): ")
if newDate != "":
    inDate = newDate.split('-')

    if len(inDate) != 3:
        print("Bad Date format")
    elif int(inDate[0]) not in range(1,13):
        print("Invalid month entered (1-12)")
    elif int(inDate[1]) not in range(1,mdays[int(inDate[0])]+1):
        print("invalid day entered (1-",mdays[int(inDate[0])],")")
    elif int(inDate[2]) not in range(21,32):
        print("invalid year entered (21-31)")
    else:
        machine.mem32[rtcBase+4] = ((2000+int(inDate[2])) << 12) | (int(inDate[0]) << 8) | int(inDate[1])
        machine.mem32[rtcBase+atomicBSet+0xc] = 0x10


# PyDOS Board Configuration for 'Adafruit ItsyBitsy RP2040 with RP2040'

Pydos_pins = {
    'sndPin' : (10,"GPIO10 D12"),
    'led' : (11,None),
    'neoPixel' : (17,None),
    'neoPixel_Pow' : (16,None),
    'I2C_NUM' : (1,None),
    'SCL' : (3,"GPIO3 SCL"),
    'SDA' : (2,"GPIO2 SDA"),
    'SPI_NUM' : (0,None),
    'SCK' : (18,"GPI18 SCK"),
    'MOSI' : (19,"GPI19 MO"),
    'MISO' : (20,"GPI20 MI"),
    'CS' : (7,"GPIO7 D9")
}



"""
    Screen/Keyboard abstraction layer

"""
from sys import stdin,stdout,implementation
if implementation.name.upper() == "MICROPYTHON":
    import uselect
elif implementation.name.upper() == "CIRCUITPYTHON":
    from supervisor import runtime

class PyDOS_UI:

    def __init__(self):
        pass

    def serial_bytes_available(self):
        if implementation.name.upper() == "CIRCUITPYTHON":
            # Does the same function as supervisor.runtime.serial_bytes_available
            retval = runtime.serial_bytes_available

        elif implementation.name.upper() == "MICROPYTHON":
            spoll = uselect.poll()
            spoll.register(stdin,uselect.POLLIN)

            retval = spoll.poll(0)
            spoll.unregister(stdin)

            if not retval:
                retval = 0

        return retval

    def read_keyboard(self,num):
        # Does the same function as sys.stdin.read(num)
        # Reads num characters from keyboard and returns
        # This is a blocking read, ie the program will wait for the input
        return stdin.read(num)

    def get_screensize(self):
        print("Screen set to 24 rows, 80 col. Press any key to coninue...",end="")
        stdout.write('\x1b[2K')
        stdout.write('\x1b[999;999H\x1b[6n')
        pos = ''
        char = ''
        try:
            char = stdin.read(1) ## expect ESC[yyy;xxxR
        except:
            return(24,80)
        if char != '\x1b':
            return(24,80)

        while char != 'R':
            pos += char
            char = stdin.read(1)
        print()

        width = int(pos.lstrip("\n\x1b[").split(';')[1],10)
        height = int(pos.lstrip("\n\x1b[").split(';')[0],10)

        if width < 1:
            width = 80
        if height < 1:
            height = 24

        return(height,width)

Pydos_ui = PyDOS_UI()



"""CircuitPython Essentials Storage logging boot.py file"""
from sys import implementation
if implementation.name.upper() == "CIRCUITPYTHON":
    import board
    import digitalio
    import storage

# Changing the value of PyDOSReadOnly from False to True changes the behavior of this program.
# When PyDOSReadOnly is False the PyDOS file system will be set to Read/Write status for next
# power cycle unless a control GPIO is grounded (GP2,5 or 6 depending on the Microcontroller)
# in which case the host computer will have Read/Write access and PyDOS will have readonly access.

    # Changing PyDOSReadOnly to True will allow the host computer write access even if the
    # GPIO pin is not grounded.

    # If a grounding Pin is not available on the microcontroller, the PyDOS file system is set to
    # ReadOnly so the host file has access unless PyDOSReadOnly is set to True in which case
    # the PyDOS file system is set to Read/Write.

    PyDOSReadOnly = False

    if board.board_id == 'arduino_nano_rp2040_connect':
# For Gemma M0, Trinket M0, Metro M0/M4 Express, ItsyBitsy M0/M4 Express
        switch = digitalio.DigitalInOut(board.D2)
    elif board.board_id in ['raspberry_pi_pico','cytron_maker_pi_rp2040']:
        switch = digitalio.DigitalInOut(board.GP6)
    elif board.board_id == 'adafruit_qtpy_esp32c3':
        switch = digitalio.DigitalInOut(board.A1)
    elif board.board_id == 'adafruit_itsybitsy_rp2040':
        switch = digitalio.DigitalInOut(board.D7)
    else:
        if "D5" in dir(board):
            switch = digitalio.DigitalInOut(board.D5)
        elif "GP5" in dir(board):
            switch = digitalio.DigitalInOut(board.GP5)
        elif "IO5" in dir(board):
            switch = digitalio.DigitalInOut(board.IO5)
        elif "D6" in dir(board):
            switch = digitalio.DigitalInOut(board.D6)
        elif "GP6" in dir(board):
            switch = digitalio.DigitalInOut(board.GP6)
        elif "IO6" in dir(board):
            switch = digitalio.DigitalInOut(board.IO6)
        elif "D2" in dir(board):
            switch = digitalio.DigitalInOut(board.D2)
        elif "GP2" in dir(board):
            switch = digitalio.DigitalInOut(board.GP2)
        elif "IO2" in dir(board):
            switch = digitalio.DigitalInOut(board.IO2)
        else:
            switch = None


    if switch:
        switch.direction = digitalio.Direction.INPUT
        switch.pull = digitalio.Pull.UP

# If the switch pin is connected to ground (switch.value == False) allow host to write to the drive
        if switch.value == False:
            # Mounts so Host computer can write to micro-flash
            storage.remount("/", True)
            print("Switch False (pin grounded), PyDOS FS is ReadOnly")
        else:
            storage.remount("/", PyDOSReadOnly)
            print("Switch True (not grounded), ",end="")
            if PyDOSReadOnly:
                print("PyDOS FS is ReadOnly")
            else:
                print("PyDOS FS is ReadWrite")

        switch.deinit()
    else:
        print("No GPIO override pin found - set to Microcontroller access mode, PyDOS FS is ReadWrite")
        print("To set the PyDOS FS readonly and give the host write access enter the following command")
        print("at the PyDOS prompt: fs ro")
        storage.remount("/", PyDOSReadOnly )
elif implementation.name.upper() == "MICROPYTHON":
    from os import uname
    if uname().machine == 'Teensy 4.1 with MIMXRT1062DVJ6A':
        import uos, sys
        uos.umount("/flash")
        uos.mount(vfs,"/")
        sys.path.pop(-1)
        sys.path.pop(-1)
        sys.path.append("/")
        sys.path.append("/lib")



#! /usr/bin/python

# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""This class defines an object that can be returned by the BASICParser to
indicate the need for a control flow branch. The information within the
object tells the Program the nature of the jump and therefore whether a
return address need to be added to the return stack.

>>> flowsignal = FlowSignal(ftype=FlowSignal.RETURN)
>>> print(flowsignal.ftarget)
-1
>>> flowsignal = FlowSignal(ftarget=100, ftype=FlowSignal.SIMPLE_JUMP)
>>> print(flowsignal.ftarget)
100
>>> print(flowsignal.ftype)
0
"""


class FlowSignal:

    # Jump categories

    # Indicates a simple jump as the result
    # of a GOTO or conditional branch. The
    # ftarget value should be the jump target, i.e.
    # the line number being jumped to
    SIMPLE_JUMP        = 0

    # Indicates a subroutine call where the
    # return address must be the line number of the instruction
    # of the following the call.
    # The ftarget value should be the line number of the first line
    # of the subroutine
    GOSUB              = 1

    # Indicates the start of a FOR loop where loop
    # variable has not reached the end value, and therefore the loop
    # must be repeated. There should be therefore be
    # no ftarget value associated with it
    LOOP_BEGIN         = 2

    # An indication from a processed NEXT statement that the loop is to
    # be repeated. Since the return address is already on the stack,
    # there does not need to be an ftarget value associated with the signal.
    LOOP_REPEAT        = 3

    # An indication from a FOR statement that the loop should be skipped because
    # loop variable has reached its end value. The ftarget should be
    # the loop variable to look for in the terminating NEXT statement
    LOOP_SKIP          = 4

    # Indicates a subroutine return has been processed, where the return
    # address is on the return stack. There should be therefore
    # be no ftarget value specified
    RETURN             = 5

    # Indicates that execution should cease because a stop statement has
    # been processed. There should be therefore be no ftarget value specified
    STOP               = 6

    # Indicates that a conditional result block should be executed
    EXECUTE            = 7

    def __init__(self, ftarget=None, ftype=SIMPLE_JUMP, floop_var=None):
        """Creates a new FlowSignal for a branch. If the jump
        target is supplied, then the branch is assumed to be
        either a GOTO or conditional branch and the type is assigned as
        SIMPLE_JUMP. If no jump_target is supplied, then a jump_type must be
        supplied, which must either be GOSUB, RETURN, LOOP_BEGIN,
        LOOP_REPEAT, LOOP_SKIP or STOP. In the latter cases
        the jump target is assigned an arbitrary value of None.

        :param ftarget: The associated value
        :param ftype: Either GOSUB, SIMPLE_JUMP, RETURN, LOOP_BEGIN,
        LOOP_SKIP or STOP
        """

        if ftype not in [self.GOSUB, self.SIMPLE_JUMP, self.LOOP_BEGIN,
                         self.LOOP_REPEAT, self.RETURN,
                         self.LOOP_SKIP, self.STOP, self.EXECUTE]:
            raise TypeError("Invalid flow signal type supplied: " + str(ftype))

        if ftarget == None and \
           ftype in [self.SIMPLE_JUMP, self.GOSUB, self.LOOP_SKIP]:
            raise TypeError("Invalid jump target supplied for flow signal type: " + str(ftarget))

        if ftarget != None and \
           ftype in [self.RETURN, self.LOOP_BEGIN, self.LOOP_REPEAT,
                     self.STOP, self.EXECUTE]:
            raise TypeError("Target wrongly supplied for flow signal " + str(ftype))

        self.ftype = ftype
        self.ftarget = ftarget
        self.floop_var = floop_var



# PyDOS Board Configuration for 'ESP32C3 module with ESP32C3' (adafruit_qtpy_esp32c3)

Pydos_pins = {
    'sndPin' : (4,"A0"),
    'neoPixel' : (2,None),
    'I2C_NUM' : (0,None),
    'SCL' : (6,"SCL"),
    'SDA' : (5,"SDA"),
    'SCK' : (10,"SCK"),
    'MOSI' : (7,"MO"),
    'MISO' : (8,"MI"),
    'CS' : (0,"A3")
}



# PyDOS Board Configuration for 'espressif_esp32s3_devkitc_1_n8r2'

import board

Pydos_pins = {
    'sndPin' : (board.IO10,"IO10"),
    'neoPixel' : (board.NEOPIXEL,None),
    'SCL' : (board.IO7,"IO7"),
    'SDA' : (board.IO6,"IO6"),
    'SCK' : (board.IO36,"IO36"),
    'MOSI' : (board.IO35,"IO35"),
    'MISO' : (board.IO37,"IO37"),
    'CS' : (board.IO9,"IO9")
}



# PyDOS Board Configuration for 'raspberry_pi_pico with Cytron Maker Pi Pico'

import cyt_mpp_board as board

Pydos_pins = {
    'sndPin' : (board.SNDPIN,"GP18 Onboard Buzzer"),
    'neoPixel' : (board.NEOPIXEL,None),
    'SCL' : (board.GP3,"GP3 GROVE#2"),
    'SDA' : (board.GP2,"GP2 GROVE#2"),
    'SD_SCK' : (board.SCK,"GP10 Onboard SD"),
    'SD_MOSI' : (board.MOSI,"GP11 Onboard SD"),
    'SD_MISO' : (board.MISO,"GP12 Onboard SD"),
    'SD_CS' : (board.GP15,"GP15 Onboard SD")
}



# PyDOS Board Configuration for 'Arduino Nano RP2040 Connect with RP2040'

Pydos_pins = {
    'sndPin' : (29,"GPIO29 A3"),
    'led' : (6,"GPIO6 D13"),
    'I2C_NUM' : (0,None),
    'SCL' : (13,"GPIO13 A5"),
    'SDA' : (12,"GPIO12 A4"),
    'SCK' : (6,"GPIO6 D13"),
    'MOSI' : (7,"GPIO7 D11"),
    'MISO' : (4,"GPIO4 D12"),
    'CS' : (5,"GPIO5 D10")
    }



# esp32s2-test.py -- small WiFi test program for ESP32-S2 CircuitPython 6
# taken from https://www.reddit.com/r/circuitpython/comments/ianpm8/using_wifi_when_running_on_esp32s2saola1_board/
#
import time
import ipaddress
import wifi
import socketpool
import ssl
import adafruit_requests
from os import getenv

# Get wifi details and more from a .env file
if getenv('CIRCUITPY_WIFI_SSID') is None:
    raise Exception("WiFi secrets are kept in .env, please add them there!")

ssid=getenv('CIRCUITPY_WIFI_SSID')
passwd=getenv('CIRCUITPY_WIFI_PASSWORD')

print('Hello World!')

for network in wifi.radio.start_scanning_networks():
    print(network, network.ssid, network.channel)
wifi.radio.stop_scanning_networks()

print("joining network...")
print(wifi.radio.connect(ssid=ssid,password=passwd))
# the above gives "ConnectionError: Unknown failure" if ssid/passwd is wrong

print("my IP addr:", wifi.radio.ipv4_address)

print("pinging 1.1.1.1...")
ip1 = ipaddress.ip_address("1.1.1.1")
print("ip1:",ip1)
print("ping:", wifi.radio.ping(ip1))


pool = socketpool.SocketPool(wifi.radio)
request = adafruit_requests.Session(pool, ssl.create_default_context())

print("Fetching wifitest.adafruit.com...");
response = request.get("http://wifitest.adafruit.com/testwifi/index.html")
print(response.status_code)
print(response.text)

print("Fetching https://httpbin.org/get...");
response = request.get("https://httpbin.org/get")
print(response.status_code)
print(response.json())

response.close()
request._free_sockets()



# PyDOS Board Configuration for 'adafruit_itsybitsy_rp2040'

import board

Pydos_pins = {
    'sndPin' : (board.D12,"GPIO10 D12"),
    'neoPixel' : (board.NEOPIXEL,None),
    'SCL' : (board.SCL,"GPIO3 SCL"),
    'SDA' : (board.SDA,"GPIO2 SDA"),
    'SCK' : (board.SCK,"GPIO18 SCK"),
    'MOSI' : (board.MOSI,"GPIO19 MO"),
    'MISO' : (board.MISO,"GPIO20 MI"),
    'CS' : (board.D9,"GPIO7 D9")
}



# PyDOS Board Configuration for 'unexpectedmaker_feathers2'

import board

Pydos_pins = {
    'sndPin' : (board.D12,"IO10 D12"),
    'dotStar_Clock' : (board.APA102_SCK,None),
    'dotStar_Data' : (board.APA102_MOSI,None),
    'SCL' : (board.SCL,"IO9 SCL"),
    'SDA' : (board.SDA,"IO8 SDA"),
    'SCK' : (board.SCK,"IO36 SCK"),
    'MOSI' : (board.MOSI,"IO35 SDO"),
    'MISO' : (board.MISO,"IO37 SDI"),
    'CS' : (board.D20,"IO33 D20")
}
